/**
@author CG
@license Adapted from https://observablehq.com/@d3/donut-chart | ISC
@image png
@chart
**/
chart.donut = function(data, {
  id,
  name = ([x]) => x,  // given d in data, returns the (ordinal) label
  value = ([, y]) => y, // given d in data, returns the (quantitative) value
  title, // given d in data, returns the title text
  width = 640, // outer width, in pixels
  height = 400, // outer height, in pixels
  innerRadius = 0, // inner radius of pie, in pixels (non-zero for donut)
  outerRadius = Math.min(width, height) / 2, // outer radius of pie, in pixels
  labelRadius = (innerRadius * 0.2 + outerRadius * 0.8), // center radius of labels
  format = ',', // a format specifier for values (in the label)
  names, // array of names (the domain of the color scale)
  colors, // array of colors for names
  stroke = innerRadius > 0 ? 'none' : 'white', // stroke separating widths
  strokeWidth = 1, // width of stroke separating wedges
  strokeLinejoin = 'round', // line join of stroke separating wedges
  padAngle = stroke === 'none' ? 1 / outerRadius : 0, // angular separation between wedges
} = {}) {
  const STYLE = (typeof AVAA_STYLE !== "undefined" ? AVAA_STYLE : null);
	
  // Compute values.
  const N = d3.map(data, name);
  const V = d3.map(data, value);
  const I = d3.range(N.length).filter(i => !isNaN(V[i]));
  //const LABELS = typeof window !== "undefined" && window.CHART_LABELS ? window.CHART_LABELS : {};
  
  // Unique the names.
  if (names === undefined) names = N;
  names = new d3.InternSet(names);

  // Chose a default color scheme based on cardinality.
  //if (colors === undefined) colors = d3.schemeSpectral[names.size];
  if (colors === undefined) colors = d3.quantize(t => d3.interpolateSpectral(t * 0.8 + 0.1), names.size);

// color scale
  const defaultColors = colors.slice();		
  const finalColors = [];
  names.forEach(ii=>finalColors.push(STYLE ? STYLE.colorOf(ii)||defaultColors.shift() : defaultColors.shift()))
	//console.info(finalColors);
	const color = d3.scaleOrdinal(names,finalColors)

  // Compute titles.
  if (title === undefined) {
    const formatValue = d3.format(format);
    //title = i => `${N[i]}\n${formatValue(V[i])}`; //
	title = i => `${STYLE ? STYLE.labelOf(N[i])||N[i] : N[i]}\n${formatValue(V[i])}`;
  } else {
    const O = d3.map(data, d => d);
    const T = title;
    title = i => T(O[i], i, data);
  }

  // Construct arcs.
  //const arcs = d3.pie().padAngle(padAngle).sort(null).value(i => V[i])(I);
  //const arc = d3.arc().innerRadius(innerRadius).outerRadius(outerRadius);
  //const arcLabel = d3.arc().innerRadius(labelRadius).outerRadius(labelRadius);

  const svg = d3.select("#"+id).attr("width", width)
      .attr("height", height)
	  .attr("viewBox", [0, 0, width, height]);
  
  const g = svg.append("g");

  g.append("g").attr("class", "slices");
  g.append("g").attr("class", "labels");
  g.append("g").attr("class", "lines");

  const radius = Math.min(width, height) / 2;
  const pie = d3
      .pie()
      .sort(null)
      .value(function(d) {
        return d.value;
      });
  
    const arc = d3
      .arc()
      .outerRadius(radius * 0.8)
      .innerRadius(radius * 0.6);
  
    const outerArc = d3
      .arc()
      .innerRadius(radius * 0.9)
      .outerRadius(radius * 0.9);
  
    g.attr("transform", "translate(" + width / 2 + "," + height / 2 + ")");
  
    const key = function(d) {
		return STYLE ? STYLE.labelOf(d.data.name) || d.data.name : d.data.name;
    };
  
  drawPie(data);
 
  function drawPie(data) {
    const pieData = pie(data);
    const slice = g
      .select(".slices")
      .selectAll("path.slice")
      .data(pieData, key);

    slice
      .enter()
      .insert("path")
      .style("fill", function(d) {
        return color(d.data.name);
      })
      .attr("class", "slice")
      .merge(slice)
      .attr("d", arc);

    slice.exit().remove();

    const labels = g
      .select(".labels")
      .selectAll("text")
      .data(pie(data), key);
  
    const _labels = labels
      .enter()
      .append("text")
      .attr("dy", ".35em")
      .text(key)
      .merge(labels)
      .attr("x", function(d) {
        const pos = outerArc.centroid(d);
        const midangle = d.startAngle + (d.endAngle - d.startAngle) / 2
        pos[0] = pos[0]  + (midangle < Math.PI ? 18 : -18);
        return  pos[0]
      })
      .attr("y", function(d) {
        const pos = outerArc.centroid(d);
        return pos[1]
      })
      .style("text-anchor", function(d) {
        var centroid = arc.centroid(d),
            midAngle = Math.atan2(centroid[1], centroid[0]),
            x = Math.cos(midAngle) * radius * 0.9;
        return x > 0? 'start' : 'end';
      });
  	
/*	opacity: .3;
	stroke: black;
	stroke-width: 2px;
	fill: none;*/
    const polyline = g
      .select(".lines")
      .selectAll("polyline")
		
      .data(pie(data), key);

    const _polyline  = polyline
      .join("polyline")
	  .style("fill","none")
	  .style("stroke","black")
	  .style("stroke-width","2px")
	  .style("opacity","0.2")
      .attr("points", function(d) {
        const theta = (d.startAngle + d.endAngle) / 2;
        const r = radius * 0.8;
        const posA = d3.pointRadial(theta,r);
        // const posA = arc.centroid(d) // line insertion in the slice
        const posB = outerArc.centroid(d) // line break: we use the other arc generator that has been built only for that
        const posC = outerArc.centroid(d); // Label position = almost the same as posB
        const midangle = d.startAngle + (d.endAngle - d.startAngle) / 2 // we need the angle to see if the X position will be at the extreme right or extreme left
        posC[0] = posC[0]  + (midangle < Math.PI ? 16 : -16);
        return [posA, posB, posC]; 
      });

    polyline.exit().remove();

    const alpha = 0.2;
    const spacing = 15;
    
    function alignLabels() {
      let again = false;
      _labels.each(function (d) {
        const a = this;
        const da = d3.select(a);
        const y1 = da.attr('y');
         _labels.each(function (d) {
          const b = this;
          const db = d3.select(b);
          if (a === b) {
              return ;
            }
           if (da.attr('text-anchor') !== db.attr('text-anchor')) {
             return ;
           }
           const y2 = db.attr('y');
           const deltaY= y1 - y2;
           
            
           if (Math.abs(deltaY) > spacing) {
             return ; 
           }
           again = true;
           const sign = deltaY > 0? 1: -1;
           const adjust = sign * alpha;
           da.attr('y', +y1 + adjust);
           db.attr('y', +y2 - adjust);
      
        })     
      })
      if(again){
        _polyline.attr("points", function(d,i) {   
          var labelForLine = d3.select(_labels.nodes()[i]);
          const theta = (d.startAngle + d.endAngle) / 2;
          const r = radius * 0.8;
          const posA = d3.pointRadial(theta,r);
          const posB = outerArc.centroid(d) 
          const posC = outerArc.centroid(d); // La
          const midangle = d.startAngle + (d.endAngle - d.startAngle) / 2 // we need the angle to see if the X position will be at the extreme right or extreme left
           posC[0] = posC[0]  + (midangle < Math.PI ? 16 : -16);
           posB[1] = labelForLine.attr('y');
           posC[1] = labelForLine.attr('y');
           return [posA, posB, posC]; 
        });
       setTimeout(alignLabels);
      }
    }
    alignLabels();
  }
  
  /*return Object.assign(svg.node(), {
    update() {
      drawPie(randomData());
    }
  });*/
  
  
  return Object.assign(svg.node(), {scales: {color}});
}