# 2024-08-13
# python -u
# Faster Whisper Wrapper for AVAA Toolkit by CG

import os, sys, argparse
import tqdm
import urllib.request 
import json
import time
import torch
from faster_whisper import WhisperModel

class _CustomProgressBarTranscribe(tqdm.tqdm):
     def __init__(self, *args, **kwargs):
          kwargs["disable"] = True
          super().__init__(*args, **kwargs)
          self._current = self.n          
          print("#P:" + str(self._current) + "/" + str(self.total))
          
     def update(self, n, count):
          #super().update(n)
          self._current += n          
          print("#P:" + str(self._current) + "/" + str(self.total)+"/"+str(count))

class _CustomProgressBarDownload(tqdm.tqdm):
     def __init__(self, *args, **kwargs):
          kwargs["disable"] = True
          super().__init__(*args, **kwargs)
          self._current = self.n
          print("#D:" + str(self._current) + "/" + str(self.total))          
          
     def update(self, n):
          #super().update(n)
          self._current += n          
          print("#D:" + str(self._current) + "/" + str(self.total))
class _NoProgressBarDownload(tqdm.tqdm):
     def __init__(self, *args, **kwargs):
          kwargs["disable"] = True
          super().__init__(*args, **kwargs)          
          #self._current = self.n
          
     def update(self, n):
          super().update(n)
          #self._current += n          

download_module = sys.modules['faster_whisper']
tqdm.auto.tqdm = _CustomProgressBarDownload
download_module.utils.tqdm = _CustomProgressBarDownload
download_module.utils.disabled_tqdm = _NoProgressBarDownload

#print(download_module.utils.disabled_tqdm);
#exit();

def main(argv):
    temperature = 0.0
    
    try:
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('-i', '--input', dest='input')
        parser.add_argument('-m', '--model', dest='model')
        parser.add_argument('-l', '--language', dest='language')
        parser.add_argument('-t', '--temperature', dest='temperature')
        parser.add_argument('-d', '--device', dest='device')
        parser.add_argument('-b', '--beam-size', dest='beamSize')
        parser.add_argument('-p', '--precision', dest='precision')
        parser.add_argument('-h', '--help', dest='help', action='store_true')
        parser.add_argument('-f', '--fake', dest='fake', action='store_true')
        parser.add_argument('-v', '--version', dest='version', action='store_true')
        parser.add_argument('-B', '--batched', dest='batched', action='store_true')      
        parser.add_argument('-V', '--verbose', dest='verbose', action='store_true')      
        args = parser.parse_args(argv)        
        
        if args.help:
            print ('avaa-faster-whisper.py -i <input> -m <model> -t <temperature>')
            sys.exit(0)
            
        if args.version:
            print ('0.1')
            sys.exit(0)
            
        if not args.input:
            print ('#E:missing input')
            sys.exit(4)           
            
        if not args.model:
            print ('#E:missing model')
            sys.exit(4)           
                
        if not args.beamSize:
            args.beamSize = "5";            
            
        if not args.precision:
            args.precision = "auto"
        
        if not args.device:
            args.device = "auto"
           
        if args.device == "auto":
            args.device = "cuda" if torch.cuda.is_available() else "cpu"
            
        if args.precision == "auto":
            args.precision = "float16" if torch.cuda.is_available() else "int8"

    except Exception as ex:
        print ('#E:'+str(ex))
        print ('#E:invalid arguments?')
        print ('#E:Got '+ (' '.join(argv)))
        sys.exit(3)    
        
            
    print ('#I:Args =', ' '.join(argv))
    print ('#I:Verbose =', 'yes' if args.verbose else 'no')
    print ('#I:Batched =', 'yes' if args.batched else 'no')
    print ('#I:Language =',  args.language or 'autodetect')
    print ('#I:Temperature =', temperature)    
    print ('#I:Beam Size =', args.beamSize)
    print ('#I:Device =', args.device)
    print ('#I:Precision =', args.precision)
    print ('#F:Input =', args.input)
    print ('#I:Model =', args.model)
    print ('#I:Loading model...')
    bench = int(time.time())
    #model = whisper.load_model(args.model)
    #args.model = "tiny"
    
    model = WhisperModel(args.model, device=args.device, compute_type=args.precision)

    print ('#I:Loaded model `'+args.model+'` in '+str(int(time.time())-bench)+' seconds')
    if not args.language:
        args.language = ''
        
    if args.language == 'autodetect': #todo
        print ('#I:Detect language:')    
        print ('#I:Detect language - Loading audio...')    
        audio = whisper.load_audio(args.input);
        # make log-Mel spectrogram and move to the same device as the model
        print ('#I:Detect language - Building spectrogram...')
        mel = whisper.log_mel_spectrogram(audio).to(model.device)        
        # detect the spoken language
        print ('#I:Detect language - Analysis...')
        try:
            _, probs = model.detect_language(mel)
            print(f"#I:Detected language: {max(probs, key=probs.get)}")
        except Exception as ex:        
            print("#W:Detection failed: "+str(ex))                    
        
    if args.fake:        
        print ('#I:All good for preview mode!')
        print ('#W:For actual transcription, use "Convert" mode.')
        sys.exit(0)        
    
    if args.language == '':
        args.language = None
    
    print ('#I:Starting transcription...')    
    if args.batched:
        print ('#I:Using batched pipeline...')
        batched_model = BatchedInferencePipeline(model=model)
        segments, info = batched_model.transcribe(args.input, task="transcribe", language=args.language, vad_filter=True, beam_size=int(args.beamSize),batch_size=16)
    else:
        segments, info = model.transcribe(args.input, task="transcribe", language=args.language, vad_filter=True, beam_size=int(args.beamSize))
    #result = model.transcribe(args.input, language=args.language, fp16=False, verbose=args.verbose if args.verbose else None, temperature=temperature)
    total_duration = round(info.duration, 2) 
    timestamps = 0.0  # to get the current segments
    count = 0
    ret = []
    with _CustomProgressBarTranscribe(total=total_duration, unit=" audio seconds") as pbar:
        for segment in segments:
            count = count + 1
            if(args.verbose):
                print ('#I:@'+str(segment.start)+': '+str(segment.text))
            pbar.update(segment.end - timestamps, count)
            timestamps = segment.end            
            ret.append({
                "start":segment.start,
                "end":segment.end,
                "text":segment.text
            })
            
        if timestamps < info.duration: # silence at the end of the audio
            pbar.update(info.duration - timestamps,count)
        #for segment in segments:
        #    print("#I:[%.2fs -> %.2fs] %s" % (segment.start, segment.end, segment.text))
    print('#R:'+json.dumps(ret))
        
if __name__ == "__main__":
    try:
        main(sys.argv[1:])
        sys.exit(0)
    except Exception as ex:        
        print("#E:"+str(ex))
        sys.exit(2)