# 2023-11-28
# python -u
# Pyannote Wrapper for AVAA Toolkit by CG
# argmax and filterwarnings inspired by https://framagit.org/justinmponcet/whisper/-/blob/main/transcribe.py
# todo https://www.gladia.io/blog/build-a-speaker-identification-system-for-online-meetings

import warnings
warnings.filterwarnings("ignore", message=".*set_audio_backend.*")

import os, sys, argparse
import urllib.request 
import json, time, base64

from pyannote.audio import Pipeline
from pyannote.core import Segment

from typing import Any, Mapping, Optional, Text

def main(argv):
    annotations = []
    try:
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('-i', '--input', dest='input')        
        parser.add_argument('-t', '--token', dest='token')
        parser.add_argument('-s', '--speakers', dest='speakers')
        parser.add_argument('-a', '--annotations', dest='annotations')
        parser.add_argument('-h', '--help', dest='help', action='store_true')
        parser.add_argument('-f', '--fake', dest='fake', action='store_true')
        parser.add_argument('-v', '--version', dest='version', action='store_true')
        parser.add_argument('-V', '--verbose', dest='verbose', action='store_true')
        args = parser.parse_args(argv)                
        if args.version:
            print ('0.1')
            sys.exit(0) 
        
        # print ('#I:ARGS =', ' '.join(argv))
        
        if not args.annotations:
            print ('#E:missing annotations');
            sys.exit(2) 
            
        if not args.token:
            print ('#E:missing hf token');
            sys.exit(2)             
            
        if not args.speakers:
            print ('#E:missing speakers count');
            sys.exit(2)                         
            
        #args.annotations = json.loads(base64.b64decode(args.annotations))
        with open(args.annotations) as f:
            args.annotations = json.load(f)                
            
    except Exception as ex:
        print ('#E:'+str(ex))
        print ('#E:invalid arguments?')
        print ('#E:Got '+ (' '.join(argv)))
        sys.exit(3)    
        
    print ('#F:Input =', args.input)
    print ('#I:Token =', args.token[:-10]+"..")    
    print ('#I:Segments = ',len(args.annotations));        
    print ('#I:Loading pyannote segmentation model...')
    bench = int(time.time());
        
    pipeline = Pipeline.from_pretrained("pyannote/speaker-diarization-3.0",
                                    use_auth_token=args.token)
                                    
    print ('#I:Loaded model in '+str(int(time.time())-bench)+' seconds')
    
                                    
    if args.fake:        
        print ('#I:All good for preview mode!')
        print ('#W:For actual diarization, use "Convert" mode.')
        sys.exit(0)        
    
    print ('#I:Starting diarization...')
    with AVAAProgressHook() as hook:    
        diarization_result = pipeline(args.input, hook=hook, num_speakers=int(args.speakers))
        
    finalAnnotations = []
    for a in args.annotations:
        segment = Segment(a['start']/1000, a['stop']/1000)
        speaker = diarization_result.crop(segment).argmax()
        finalAnnotations.append({
            "start":a['start'],
            "stop":a['stop'],            
            "tier":speaker
        })        
            
    print ('#R:'+json.dumps(finalAnnotations))
    sys.exit(0) 

class AVAAProgressHook:
    """Hook to print progress 
    inspired by https://github.com/pyannote/pyannote-audio/blob/develop/pyannote/audio/pipelines/utils/hook.py        
    """
    def __init__(self):                
        return None

    def __enter__(self):                
        return self

    def __exit__(self, *args):        
        return None

    def __call__(
        self,
        step_name: Text,
        step_artifact: Any,
        file: Optional[Mapping] = None,
        total: Optional[int] = None,
        completed: Optional[int] = None,
    ):        
        if completed is None:
            completed = total = 1

        if not hasattr(self, "step_name") or step_name != self.step_name:
            self.step_name = step_name
            #print("#I:"+step_name+"...")            

        print("#P:"+str(completed)+"/"+str(total)+"/"+self.step_name)
        #self.progress.update(self.step, completed=completed, total=total)
        
        
if __name__ == "__main__":
    try:
        main(sys.argv[1:])
        sys.exit(0)
    except Exception as ex:        
        print("#E:"+str(ex))
        sys.exit(2)