# 2023-11-28
# python -u
# Whisper Wrapper for AVAA Toolkit by CG
# Credits:
#    - Custom Progress Bar hook concept by aadnk https://github.com/openai/whisper/discussions/850
# Todo:
#    - https://community.openai.com/t/whisper-hallucination-how-to-recognize-and-solve/218307/16

import os, sys, argparse
import tqdm
import urllib.request 
import whisper.transcribe 
import whisper
import json
import time

class _CustomProgressBarTranscribe(tqdm.tqdm):
     def __init__(self, *args, **kwargs):
          super().__init__(*args, **kwargs)
          self._current = self.n          
          
     def update(self, n):
          #super().update(n)
          self._current += n          
          print("#P:" + str(self._current) + "/" + str(self.total))

class _CustomProgressBarDownload(tqdm.tqdm):
     def __init__(self, *args, **kwargs):
          super().__init__(*args, **kwargs)
          self._current = self.n
          
     def update(self, n):
          #super().update(n)
          self._current += n          
          print("#D:" + str(self._current) + "/" + str(self.total))
          

# Inject into Whisper, so we can bubble the progress
transcribe_module = sys.modules['whisper.transcribe']
transcribe_module.tqdm.tqdm = _CustomProgressBarTranscribe
download_module = sys.modules['whisper']
download_module.tqdm = _CustomProgressBarDownload

def main(argv):
    temperature = 0.0
    
    try:
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('-i', '--input', dest='input')
        parser.add_argument('-m', '--model', dest='model')
        parser.add_argument('-l', '--language', dest='language')
        parser.add_argument('-t', '--temperature', dest='temperature')
        parser.add_argument('-h', '--help', dest='help', action='store_true')
        parser.add_argument('-f', '--fake', dest='fake', action='store_true')
        parser.add_argument('-v', '--version', dest='version', action='store_true')
        parser.add_argument('-V', '--verbose', dest='verbose', action='store_true')      
        args = parser.parse_args(argv)        
        
        if args.help:
            print ('avaa-whisper.py -i <input> -m <model> -t <temperature>')
            sys.exit(0)
            
        if args.version:
            print ('0.1')
            sys.exit(0)
            
        if not args.input:
            print ('#E:missing input')
            sys.exit(4)           
            
        if not args.model:
            print ('#E:missing model')
            sys.exit(4)           
                
        
    except Exception as ex:
        print ('#E:'+str(ex))
        print ('#E:invalid arguments?')
        print ('#E:Got '+ (' '.join(argv)))
        sys.exit(3)    
        
            
    print ('#I:Args =', ' '.join(argv));
    print ('#I:Verbose =', 'yes' if args.verbose else 'no');
    print ('#F:Input =', args.input)
    print ('#I:Model =', args.model)
    print ('#I:Language =',  args.language or 'autodetect')
    print ('#I:Temperature =', temperature)
    print ('#I:Loading model...')
    bench = int(time.time());
    model = whisper.load_model(args.model)
    print ('#I:Loaded model `'+args.model+'` in '+str(int(time.time())-bench)+' seconds')
    if not args.language:
        args.language = ''
        
    if args.language == '':
        print ('#I:Detect language:')    
        print ('#I:Detect language - Loading audio...')    
        audio = whisper.load_audio(args.input);
        # make log-Mel spectrogram and move to the same device as the model
        print ('#I:Detect language - Building spectrogram...')
        mel = whisper.log_mel_spectrogram(audio).to(model.device)        
        # detect the spoken language
        print ('#I:Detect language - Analysis...')
        try:
            _, probs = model.detect_language(mel)
            print(f"#I:Detected language: {max(probs, key=probs.get)}")
        except Exception as ex:        
            print("#W:Detection failed: "+str(ex))                    
        
    if args.fake:        
        print ('#I:All good for preview mode!')
        print ('#W:For actual transcription, use "Convert" mode.')
        sys.exit(0)        
    
    if args.language == '':
        args.language = None
    
    print ('#I:Starting transcription...')    
    result = model.transcribe(args.input, language=args.language, fp16=False, verbose=args.verbose if args.verbose else None, temperature=temperature)
  
    print('#R:'+json.dumps(result['segments']))
        
if __name__ == "__main__":
    try:
        main(sys.argv[1:])
        sys.exit(0)
    except Exception as ex:        
        print("#E:"+str(ex))
        sys.exit(2)