/** 
@author CG
@log 2025-01-07 initial implementation
@input Array<Annotation>
@output Array<Annotation>
@novalue

@attr vocabulary string () The Controlled Vocabulary ID to use for checking the value of the annotations
@attr fix-whitespaces bool (false) Whether to correct when mistake is surrounding whitespaces
@attr fix-repeats bool (false) Whether to correct when mistake is a repeated value
@attr return select (input) [input,valid,invalid,corrected] The operation by default returns the original (eventually corrected) input, but can also return the list of valid (or invalid) annotations

Checks each annotation value against a specified controlled vocabulary. 
If the value does not match a controlled vocabulary item, a warning will be issued.
The operation can also fix common mistakes, in which case the original input annotations will be modified (use the "clone" operation beforehand to prevent original annotations modification).

@dependency group-annotations.js
@dependency sort-annotations.js

**/
function(attribute, input){	
	if(!attr.vocabulary){
		log.error(`you must specify a vocabulary ID`)
		return input;
	}
	if(!Array.isArray(input)){		
		log.error(`input must be an array`);
		return input;
	}
	const returnMode = attr['return'] || 'input';
	const fixRepeats = attr['fix-repeats']==='true';
	const fixWhitespaces = attr['fix-whitespaces']==='true';
	const voc = input[0].af.vocabulary.values().toArray().filter(cve=>cve.cv==attr.vocabulary);
	if(!voc.length){
		log.error(`can't find a Controlled Vocabulary with id "${attr.vocabulary}"`);
		return input;
	}
	log.info(`Controlled Vocabulary "${attr.vocabulary}" has ${voc.length} items`);
	const controlledValues = voc.map(cve=>cve.value);
	let invalid = [], valid = [], fixed = [];
	for(let a of input){
		if(controlledValues.includes(a.value)){
			valid.push(a);
			continue;
		}
		invalid.push(a);
		log.info(a.value);
		if(fixRepeats){ //todo fix this naive fixing method
			let found = [];
			for(let cvi of controlledValues){
				if(a.value.includes(cvi)) found.push(cvi);
			}
			if(found.length===1){
				a.value = found[0];
				fixed.push(a);
				continue;
			}
		}
		if(fixWhitespaces){
			if(controlledValues.includes(a.value.trim())){
				a.value = a.value.trim();
				fixed.push(a);
				continue;
			}
		}
	}
	if(invalid.length) log.warn(`${invalid.length} invalid annotations`);
	if(fixed.length) log.warn(`${invalid.length} corrected annotations`);
	switch(attr['return']){
		case 'valid': return valid;
		case 'invalid': return invalid;
		case 'corrected': return fixed;
	}
	return input;
}