/** 
@author CG
@log 2024-08-14 added match-to-original-annotations
@log 2023-03-20 initial implementation
@input None
@output Array<Annotation>

@attr dimension string The dimension to extract from the form
@attr exclude-from-intercoding bool (false) Whether to tag the extracted annotations so they are excluded from intercoding calculations
@attr extract-original-annotations bool (false) Whether to extract only the original annotations from the form, so the results are safe from corpus changes. When this attribute is specified, no dimension will be extracted: use again the operation to extract a dimension.
@attr match-to-original-annotations bool (false) Try to correct loaded annotations so they all have a match with the original ones

Loads JSON files resulting from forms, as virtual annotations.
The JSON files must be in the folder of the processed XML file.

@sample OP load-annotations-from-form="my form (*).json" 
**/
function(filter, v , c){
	const files = api.listFiles(filter);
	if(!files.length){
		log.error(`No files found with filter "${filter}"`);
		return `data;`
	}	
	const matchToOriginalAnnotations = attr['match-to-original-annotations'] == 'true';
	const excludeFromIntercoding = attr['exclude-from-intercoding'] == 'true';
	const extractOriginalAnnotations = attr['extract-original-annotations'] == 'true';
	const addTag = attr['add-tag'] ? attr['add-tag']+'' : '';
	attr.dimension = attr.dimension ? attr.dimension+'':'';	
	const tiers = [];
	const finalTiers = [];
	const dimensions = {};		
	const cleanFileName = s=>{
		s=(s+'').split('\\').pop().split('/').pop();
		filter.split('*').forEach(p=>s=s.split(p).join(''));
		return s;
	}
	const originalAnnotations = [];	
	const ret = [];
	if(extractOriginalAnnotations || matchToOriginalAnnotations){		
		var listCount=-1;
		var listFile;		
		files.forEach(f=>{
			let json = JSON.parse(api.readFile(f));		;
			if(listCount<0){
				listCount = json.list.length;
				listFile=f;				
			}			
			else if(listCount!=json.list.length){
				log.warn(`Forms don't contain same amount of results: ${cleanFileName(f)}=${json.list.length} ${cleanFileName(listFile)}=${listCount}`);
			}
			if(listCount<json.list.length){ // we keep the one with most number of results
				listCount = json.list.length;
				listFile = f;
			}
		});
		let f = listFile || files[0]; //TODO: from a master form file?
		let json = JSON.parse(api.readFile(f));		;
		json.list.forEach(o=>{
			let afid = o.file || o.af || o.eaf; //legacy
			let af = api.getAF(afid);
			if(!af) return log.warn(`can't find annotation file "${afid}" in corpus`);
			//log.info(af.id);
			let tier = af.getTier(o.tier);
			if(!tier) tier = af.addTier({id:o.tier});
			let a = af.addAnnotation({ //TODO .createAnnotation so it doesn't affect current domain state?
				tier:tier,
				id:o.tier+"_"+o.start+"_"+o.stop, //TODO harmonize
				start:o.start,
				stop:o.stop,
				value:o.value
			});				
			if(excludeFromIntercoding) a.addTag('exclude-from-intercoding');
			originalAnnotations.push(a);
		})
		if(extractOriginalAnnotations){
			if(originalAnnotations.length) log.info(`extracted ${originalAnnotations.length} original annotations from ${cleanFileName(f.name)}`);
			else log.warn(`no original annotation found!`)			
			return originalAnnotations;
		}		
	}
	files.forEach(f=>{		
		var tierName = (f.getName()+"").split("(")[1].split(')')[0];
		tiers.push(tierName);
		finalTiers.push(tierName);
		log.debug('Loading '+tierName);
		try {
			var dim = {};
			var json = JSON.parse(api.readFile(f));
			if(!json.representation){
				json.representation = {"dimensions":[{"name":"choix","label":"Choisissez","type":"select","choices":"chef,map,locX,map-et-loc,chef-et-map,chef-et-loc"},{"name":"justification","label":"Justifiez","type":"text"}],"fields":[{"name":"nom","label":"Entrez votre nom","type":"text"}]};
				json.list.forEach(o=>{
					o.dimensions = {choix:o.input, justification:''}
				});
				log.warn('used special clo-fix');
			}
			if(json.representation && json.representation.dimensions){
				json.representation.dimensions.forEach(d=>{
					dimensions[d.name] = d;
					dim[d.name] = d;
				})
			}			
			
				
			
			
			if(!attr.dimension){
				attr.dimension = Object.keys(dimensions)[0];
				if(!attr.dimension){
					log.error(`no dimension found to use as default dimension`);
					return;
				}else{
					log.info(`using dimension "${attr.dimension}" because none was specified`);
				}
			}				
			
			if(attr.dimension.charAt(0)=='~'){
				//attr.dimension = attr.dimension.
			}
			if(!dim[attr.dimension]){
				log.warn(`dimension "${attr.dimension}" not found in form of tier "${tierName}"`);
				return;
			}
			json.list.forEach(o=>{
				var afid = o.file || o.af || o.eaf; //legacy
				var af = api.getAF(afid);
				if(!af){
					log.warn(`can't find annotation file "${afid}" in corpus`);
					return;
				}
				let T = tierName+'';
				if(attr.dimension) T = T + ' - ' +attr.dimension;
				//log.warn(T);
				if(finalTiers.indexOf(T)==-1) finalTiers.push(T);
				let tier = af.getTier(T);
				if(!tier) tier = af.addTier({id:T});
				
				let value = o.dimensions ? o.dimensions[attr.dimension] : o.input;
				let a = af.addAnnotation({
					tier:tier,
					id:"",
					start:o.start,
					stop:o.stop,
					value:value
				});
				a.setExtra('form',true);
				if(excludeFromIntercoding) a.addTag('exclude-from-intercoding');
				ret.push(a);
			});
		}catch(ex){
			log.error(`FAILED TO READ FILE: ${f.getName()}`);			
			log.error(ex.message);
		}
	});
	log.info('found tiers: '+tiers.join(', '));
	log.info('found dimensions: '+Object.keys(dimensions).join(', '));
	
	
	
	const findClosestWithSameMS = (target,list)=>{
		let closest;
		let diff;
		for(let a of list){
			if(a.af!=target.af) continue;
			let d1 = String(a.stop-a.start).slice(-3); // get milliseconds part
			let d2 = String(target.stop - target.start).slice(-3);			
			if(d1!=d2) continue; // this is a fix for old extend-duration issue in form
			let d = Math.abs(target.start-a.start);
			if(!closest || diff>d){
				closest=a;
				diff=d;
			}
		}
		return closest;
	}
	
	const findClosest = (target,list)=>{
		let closest = findClosestWithSameMS(target,list);
		if(closest) return closest;
		let diff;
		for(let a of list){
			if(a.af!=target.af) continue;			
			let d = Math.abs(target.start-a.start);
			if(!closest || diff>d){
				closest=a;
				diff=d;
			}
		}
		return closest;
	}
	
	if(matchToOriginalAnnotations){
		const fixReport = {annotations:[]}
		if(!originalAnnotations.length){
			log.warn(`match-to-original-annotations: no original annotations were found`);
		}else{
			let oStart = {}
			for(let a of originalAnnotations){
				oStart[a.start] = a;
			}			
			for(let a of ret){
				if(!oStart[a.start]){
					fixReport.annotations.push(a.toString());
					let closest = findClosest(a,originalAnnotations);
					if(!closest){
						log.warn(`failed to find closest annotation for ${a}`);
					}else{
						a.start = closest.start;
						a.stop = closest.stop;
					}
				}
			}
		}
		if(fixReport.annotations.length){			
			log.warn(`corrected ${fixReport.annotations.length} annotations to match originals`)
			log.report(fixReport);
		}
	}
	
	log.info(`extracted ${ret.length} results as new annotations`);
	return ret;
}	
