/** 
@author CG
@license Speech Synthesizer https://github.com/kripken/speak.js | GPL-3.0
@license Original eSpeak library ported by speak.js | GNU GPL
@log 2024-03-21 migration to pipe temp
@log 2023-12-08 migration to ffmpeg-wrapper
@log 2023-11-27 migration to processor 
@log 2023-07-06 fix amplitude issues in filters + add progress indicator
@input Array<Annotation>

@attr mode string (silence) [silence,noise,beep,voice,file] The mode of anonymization, an audio transformation that will be applied to each annotation segment.
@attr mute bool (true) Whether to silence the segment anyway before mixing the noise/beep/voice/file
@deprec @attr output string Path to save anonymized video (default: original name + suffix)
@deprec @attr suffix string (_anon) A name appended to the output (anonymized) video file
@attr beep-frequency int (800) The frequency to use for sine beep sound
@todo @attr beep-weight number (1) Mixing weight of the sine beep sound
@attr noise-amplitude number (1) [0 1] Amplitude of the generated noise
@attr noise-color string (white) [white,pink,brown,blue,violet,velvet] Noise color
@attr noise-seed int Seed value for noise PRNG
@attr noise-weight number (1) Mixing weight of the noise sound
@attr voice-text string Text to synthetize and use for anonymization. If not provided, the annotation's value will be used.
@attr voice-amplitude int (100) How loud the voice will be.
@attr voice-pitch int (50) The voice pitch
@attr voice-speed int (175) The speed at which to talk (words per minute)
@todo voice-type string Which voice to use (for a non-default voice, requires you to build speak.js to include the proper data. See Language Support below) (default: en/en-us)
@attr voice-wordgap int (0) Additional gap between words in 10 ms units
@attr file string Path to an audio file to use for anonymization, when mode is set to "file"
@deprec @attr start number Optional start time (in seconds)
@deprec @attr stop number Optional stop/end time to cut clip at (in seconds)
@param annotation-value="X" file="X.mp3" | Use a custom audio file for annotations with a specific value
@param annotation-file="X" output-file="Y.mp4" | Map custom output filename for a given annotation file
@param !file
@param !output-file

Audio Anonymizer modifies media files by applying audio filters on each input annotation segment.
Available modes are
- silence: replaces each segment with complete silence (default)
- noise: replaces each segment with a configurable noise
- beep: replaces each segment with a configurable beep
- voice: replaces each segment with a synthetized voice
- file: replaces each segment with a custom audio file

@dependency proc-ffmpeg-wrapper.js
@dependency speakGenerator.js
@todo refactor
**/
proc['audio-anonymizer'] = {	
	pipe:pipe=>{		
		if(!data || !data.length || !data[0].af){
			log.error(`audio anonymizer expects an array of annotations as input`);
			return false;
		}
		const list = data;
		const mediaStart = attr.start*1 || 0;
		const mediaStop = attr.stop*1 || 0;
		const mode = attr.mode || 'silence';
		const mute = attr.mute=='false' ? false : true;
		const beepFrequency = attr['beep-frequency']*1 || 800;
		const beepWeight = attr['beep-weight']*1 || 1;
		const noiseAmplitude = attr['noise-amplitude']*1 || 1;
		const noiseSeed = attr['noise-seed']*1 || 0;
		const noiseColor = attr['noise-color'] || 'white';
		const noiseWeight = attr['noise-weight']*1 || 1;
		const mixFile = attr.file || '';
		const voiceText = attr['voice-text'] || '';
		const voiceOptions = {
			noWorker:true,
			amplitude: attr['voice-amplitude']*1 || 100,
			pitch: attr['voice-pitch']*1 || 100,
			speed: attr['voice-speed']*1 || 175,
			wordgap: attr['voice-wordgap']*1 || 0
		};		
		
		var isTest = api.isPreviewMode();
		var isEstimateMode = api.isEstimateMode();
		if(isTest){
			log.info(`Skipped real anonymisation because preview mode!`);			
		}			
		var voices = {};
		var files = {};
		var outputs = {};
		params.forEach(opt=>{ 
			var o = {};
			for(var i = 0; i < opt.getAttributes().getLength();i++){
                var att = opt.getAttributes().item(i).getNodeName()+'';
				o[att] = opt.getAttributes().item(i).getTextContent()+'';
			}				
			if(o['annotation-value']){
				files[o['annotation-value']]=o['file'];
			}
			if(o['annotation-file']){
				outputs[o['annotation-file']]=o['output-file'];
			}
		});	
		//log.debug("files = "+JSON.stringify(files));
		if(mode=='file' && !mixFile && !Object.keys(files).length){
			log.error(`No audio file was set for this anonymizing mode. You must set the "file" attribute or add "file" parameters`);
			return false;
		}
		
		var asort = (A,B)=>{ return A.start > B.start ? 1 : -1};
		var eafs = {}; 
		list.forEach(a=>{
			if(!a.af) throw `input is not entirely an array of annotations`;
			if(!eafs[a.af.id]) eafs[a.af.id] = [];
			eafs[a.af.id].push(a);			
		});
		var benchAll = Date.now();				
		
		for(var eafid in eafs){	
			var af = "", cf = ""; // audioFilter / complexFilter			
			var bench = Date.now();					
			let f = pipe.getCorpusFile(eafid);			
			if(!f){
				// probably corpus got reduced
				// TODO refactor
				continue;
			}
			var step = log.addStep({name:`${eafid}`,state:'pending'});			
			if(!f.media.size()){
				step.error(`No media file available in ${eafid}`);		
				continue;
			}
			for(let pmf of f.media){						
				step.file(pmf,`Input = `);
				if(!pmf.file.exists()){
					step.error(`The input file does not exist!`)
					continue;
				}
				let mediaDuration = pmf.mediaFile.getDetails().duration;
				let mediaFormat = pmf.mediaFile.getDetails().format;				
				if(!mediaDuration){
					step.error(`Invalid duration. Is the media file corrupted?`)
					continue;
				}
				/*var output = outputs[eafid]||''; //attr.output;
				if(!output){				
					var suffix = attr.suffix || '_anon';
					var name = pmf.mediaFile.name;
					if(name.indexOf('.')<0) output=pmf.mediaFile.folder+"/"+name+suffix;
					else {
						name = name.split('.');
						name[name.length-2] = name[name.length-2]+suffix;
						output=pmf.mediaFile.folder+"/"+name.join('.');
					}				
				}else{
					if(output.indexOf('/')<0 && output.indexOf('\\')<0){
						output = pmf.mediaFile.folder+'/'+output;
					}
				}*/			
				let output = pipe.temp(pmf);
				step.file(output, `Output = `);
				//step.info(`Duration = ${mediaDuration}s [${mediaFormat}]`);
				let outFile = new java.io.File(output);
				if(outFile.exists()) step.warn(`The file already exists and will be overwritten`)				
				//mediaFile.toLowerCase().replace(".mp4","_ANON.mp4")
				var inputs = []; 
				var eaf = eafs[eafid][0].af			
				eafs[eafid].sort(asort);
				
				// OVERLAP CHECK			
				var merged = [];
				eafs[eafid].forEach((a,i)=>{
					merged.push(a);
					if(merged.length<2) return;
					var p = merged[merged.length-2];
					if(p.stop>a.start){
						let mergedAnnotation = String(merged.pop());
						step.warn(`Segments are overlaping and got merged around (${mergedAnnotation})`);						
						p.stop=a.stop;					
					}				
				});
				let base = `Anonymisation of ${merged.length} segment${merged.length>1?'s':''}... ` 
				step.info(base);
				merged.forEach((a,i)=>{										
					var start = a.start/1000;
					var stop = a.stop/1000;
					//start = Math.round(start * 10) / 10;
					//stop = Math.round(stop * 10) / 10;					
					if(mediaStop && start>mediaStop) return; // skip if not in range
					var d = stop - start;					
					switch(mode){
						case 'mute':
						case 'silence':
							if(af) af+=', ';
							af += `volume=enable='between(t,${start},${stop})':volume=0`;						
							break;
						case 'beep':
							if(af) af+=', ';
							var weights = mute ? `1 1` : `1 ${beepWeight}`;
							af += `volume=enable='between(t,${start},${stop})':volume=${mute?0:1}[main];sine=d=${d}:f=${beepFrequency},adelay=${Math.floor(start*1000)},pan=stereo|FL=c0|FR=c0[beep];[main][beep]amix=inputs=2:normalize=0`; //:weights='${weights}',volume=2						
							break;
						case 'noise':
							if(af) af+=', ';
							var weights = mute ? `1 1` : `1 ${noiseWeight}`;						
							var seed = noiseSeed ? `:s=${noiseSeed}` : '';
							af += `volume=enable='between(t,${start},${stop})':volume=${mute?0:1}[main];anoisesrc=d=${d}:c=${noiseColor}:a=${noiseAmplitude}${seed},adelay=${Math.floor(start*1000)},pan=stereo|FL=c0|FR=c0[noise];[main][noise]amix=inputs=2:normalize=0`;						
							break;	
						case 'voice':
							var text = voiceText ? voiceText : a.value;
							text+=''; // still needed since primitivewrapfactory?
							if(!text){
								step.warn(`text for speech is empty in annotation: `+a.toString());
								break;
							}						
							if(!voices[text]){
								step.info(`Generating speech: ${text}`);						
								step.work();
								var wav = generateSpeech(text, voiceOptions);
								var file = api.writeBase64File("speak"+i+".wav",encode64(wav));
								voices[text] = file.getAbsolutePath();						
								step.ok();
							}
							inputs.push(voices[text]);
							var I = inputs.length;
							cf += cf ? `[main${I}]` : `[0:a]`;
							cf += `volume=enable='between(t,${start},${stop})':volume=${mute?0:1}[main${I+1}];`
							cf += `[${I}:a]adelay=${Math.floor(start*1000)}|${Math.floor(start*1000)}[aud${I}];`
							break;
						case 'file':
							var input = files[a.value+''] ? files[a.value+''] : mixFile;
							if(!input){
								step.warn(`No file specified to replace segment`)
								break;
							}
							inputs.push(input);
							var I = inputs.length;					
							cf += cf ? `[main${I}]` : `[0:a]`;
							cf += `volume=enable='between(t,${start},${stop})':volume=${mute?0:1}[main${I+1}];`
							cf += `[${I}:a]adelay=${Math.floor(start*1000)}|${Math.floor(start*1000)}[aud${I}];`
							//report+=`<td>${input}</td>`;
							break;
					}
				});
				
				if(mode=='voice' || mode=='file'){
					cf += `[main${inputs.length+1}]`;
					for(var i=0;i<inputs.length;i++) cf+=`[aud${i+1}]`;
					cf += `amix=inputs=${inputs.length+1}:duration=first:normalize=0`;	//dropout_transition=0:
				}
				//if(af) step.debug('AF='+af);
				//if(cf) step.debug('CF='+cf);			
				var o = {			
					output:output,
					vcodec:'copy',
					start:mediaStart*1000			
				}
				if(af) o.af = `"${af}"`;
				if(cf) o.cf = `"${cf}"`;
				if(inputs.length) o.inputs = inputs;
				if(mediaStop) o.stop = mediaStop*1000;				
				var durationMS = Math.round(pmf.mediaFile.getDetails().duration*1000);
				// when using media.clip(), we must handle progress ourselves
				if(isTest || isEstimateMode) {
					// in test mode, converter expects we provide expected progress duration				
					api.converter.progressTodo(durationMS);
					api.sleep(500);
					step.replace(`${base} skipped because preview mode`);
				}else {				
					let ffargs = ['-y',
						'-i',pmf.getAbsolutePath(),
						'-vcodec','copy'
					]
					if(af) ffargs.push('-af','"'+af+'"');
					if(cf) ffargs.push('-cf','"'+cf+'"');			
					ffargs.push(output)					
					let rs = ffmpegWrapper(step,ffargs,{
						onProgress:o=>{
							// we must inform about the current progress
							api.converter.progressDone(o.chunk);
						}
					});
					if(!rs) return false; // abort pipe
					pmf.replaceWith(output);					
				}				
			}
		}		
		if(!isTest){
			log.info(`Anonymization completed in ${Math.ceil((Date.now()-benchAll)/1000)} seconds`)
		}		
	}
}


function parseWav(wav) {
	function readInt(i, bytes) {
	  var ret = 0;
	  var shft = 0;
	  while (bytes) {
		ret += wav[i] << shft;
		shft += 8;
		i++;
		bytes--;
	  }
	  return ret;
	}
	if (readInt(20, 2) != 1) throw 'Invalid compression code, not PCM';
	if (readInt(22, 2) != 1) throw 'Invalid number of channels, not 1';
	return {
	  sampleRate: readInt(24, 4),
	  bitsPerSample: readInt(34, 2),
	  samples: wav.subarray(44)
	};
}
function encode64(data) {
     var BASE = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
     var PAD = '=';
     var ret = '';
     var leftchar = 0;
     var leftbits = 0;
     for (var i = 0; i < data.length; i++) {
        leftchar = (leftchar << 8) | data[i];
        leftbits += 8;
        while (leftbits >= 6) {
          var curr = (leftchar >> (leftbits-6)) & 0x3f;
          leftbits -= 6;
          ret += BASE[curr];
        }
      }
      if (leftbits == 2) {
        ret += BASE[(leftchar&3) << 4];
        ret += PAD + PAD;
      } else if (leftbits == 4) {
        ret += BASE[(leftchar&0xf) << 2];
        ret += PAD;
      }
	return ret;
}