/** 
@author CG
@log 2023-12-14 rewrote
@log 2023-11-30 initial implementation
@cachable

@input Array<Annotation>

@attr start string () Starting point in seconds to start cutting from
@attr duration string () Duration of cut segment
@attr concat bool (false) Whether to concat all segments in case multiple are cut

@dependency milliseconds-to-time.js
@dependency proc-ffmpeg-wrapper.js

Cuts a segment from each corpus media file.

This processor also accepts an array of annotations to cut multiple segments. In this case, the corpus will be reduced to relevant annotation files and each media file will be replaced by its cuts, or by a one merged file from all cuts when "concat" attribute is set to true.

**/
proc['ffmpeg-cut'] = {	
	pipe:pipe=>{
		const attrConcat = attr.concat=='true';		
		
		if(data && data.length){			
			log.info(`Using annotations as segments to cut`);
			const asort = (A,B)=>{ return A.start > B.start ? 1 : -1};
			let domain = {}; 			
			data.forEach(a=>{
				if(!domain[a.af.id]) domain[a.af.id] = {list:[],af:a.af,media:a.af.getMedia()};
				domain[a.af.id].list.push(a);				
			});			
			const domainFilesToKeep = [];
			for(let afid in domain){
				let list = domain[afid].list;
				let f = pipe.getDomainFile(afid); // we want to work on pipelined files				
				domainFilesToKeep.push(f);
				let step = log.addStep({name:afid});				
				if(!f.media.size()) {
					step.error(`No media file available in ${afid}`);
					continue;
				}			
				let created = [];
				for(let pmf of f.media){
					if(api.isEstimateMode()){
						api.converter.progressTodo(list.reduce((total, annotation) => total + annotation.getDuration(),0));
						continue;
					}
					step.file(pmf,`Input = `);					
					if(api.isPreviewMode()){
						step.info(`Not cutting because preview mode.`)
						continue;
					}				
					let concatList = [];
					for(a of list){
						let cutFile = pipe.temp(pmf);						
						step.info(`Segment from ${msToTime(a.start)} to ${msToTime(a.stop)}`); 
						step.file(cutFile,`Output = `);
						let ffargs = [
							'-ss',a.start/1000,
							'-i',pmf.file.path,
							'-c','copy',							
							'-t',(a.stop-a.start)/1000,
							cutFile.getAbsolutePath()
						]			
						let rs = ffmpegWrapper(step,ffargs,{
							expectedDuration:a.getDuration(),
							onProgress:o=>api.converter.progressDone(o.chunk)
						});
						if(!rs) pipe.abort();
						created.push(cutFile);
						concatList.push(cutFile.getAbsolutePath());						
					};
					if(attrConcat){
						if(concatList.length){
							let concatOut = pipe.temp(pmf);
							let concatIn = pipe.temp("concat.txt")
							step.file(concatOut,`Joining ${concatList.length} segments into `);
							api.writeFile(concatIn, concatList.map(s=>`file '${s}'`).join("\n"))													
							let ffargs = [
								'-safe','0',
								'-f','concat',
								'-i',concatIn.getAbsolutePath(),
								'-c','copy',							
								concatOut.getAbsolutePath()
							]											
							let rs = ffmpegWrapper(step,ffargs,{
								expectedDuration:a.getDuration(),
								onProgress:o=>api.converter.progressDone(o.chunk)
							});
							if(!rs) pipe.abort();
							pmf.replaceWith(concatOut);								
						}
					}
				} // end of media
				if(!attrConcat){
					f.media.clear();
					for(mf of created) f.addMedia(mf);
				}
			} // end of domain			
			// now let's keep in the pipeline only processed files
			pipe.domain.clear();
			for(let df of domainFilesToKeep) pipe.domain.add(df);			
		}else{			
			if(!attr.start || !attr.duration){
				log.error(`you must specify start and duration attributes, or provide a selection of annotations`);
				return false;
			}
			if(!attr.duration) throw `you must specify duration attribute`;
			for(let f of pipe.domain){					
				let step = log.addStep({name:f.id});
				if(!f.media.size()){
					step.error(`No media file available in ${f.id}`);
					continue;
				}				
				for(let pmf of f.media){						
					if(api.isEstimateMode()) api.converter.progressTodo(attr.duration*1000);				
					step.file(pmf,`Input = `);
					if(!pmf.exists()){
						step.error(`The file does not exist.`);
						continue;
					}
					if(api.isPreviewMode()){
						step.info(`Not cutting because preview mode.`)
						continue;
					}				
					let cutFile = pipe.temp(pmf);
					step.file(cutFile,`Output = `);
					let ffargs = [
						'-ss',attr.start,
						'-i',pmf.getAbsolutePath(),
						//'-vf','scale=720:480',
						'-c','copy',
						'-t',attr.duration,
						cutFile.getAbsolutePath()
					];
					const rs = ffmpegWrapper(step,ffargs,{
						expectedDuration:attr.duration*1000,
						//trap:s=>step.debug(s),
						onProgress:o=>api.converter.progressDone(o.chunk)
					});
					if(!rs) pipe.abort();
					// can't cut images for instance so sometimes it breaks?
					if(cutFile.exists()) pmf.replaceWith(cutFile); // update pipe media		
				}			
			}
		}
	}
}
