/** 
@author CG
@log 2024-08-14 added RNN and NLM modes
@log 2023-12-04 initial implementation
@license Noise Removal Neural Network Models https://github.com/GregorR/rnnoise-models | No

@experimental
@cachable

@attr method string () [FFT,RNN,NLM] The denoise method to use
@attr rnn-model string (beguiling-drafter) [beguiling-drafter,conjoined-burgers,leavened-quisling,marathon-prescription,somnolent-hogwash,xiph-std] 
@attr rnn-mix number (1) How much to mix filtered samples into final output. Allowed range is from -1 to 1. Default value is 1. Negative values are special, they set how much to keep filtered noise in the final filter output
@attr rnn-threads int (1) Number of threads (1 for stereo)
@attr nlm-strength number (0.00001) [0.00001,10000] Set denoising strength. Allowed range is from 0.00001 to 10000
@attr nlm-patch number (0.002) [0.001,0.1] Set patch radius duration. Allowed range is from 0.001 to 0.1
@attr nlm-research number (0.006) [0.002,0.3] Set research radius duration. Allowed range is from 2 to 300 milliseconds
@attr nlm-smooth number (11) [1,1000] Set smooth factor. Allowed range is from 1 to 1000
@attr nlm-output select () [o] [output denoised,input unchanged,noise only] Set the output mode

This processor calls ffmpeg's denoise feature.
- FFT: Denoises audio with FFT.
- NLM: Reduces broadband noise using a Non-Local Means algorithm.
- RNN: Reduces noise from speech using Recurrent Neural Networks model.

Learn more about the RNN models at https://github.com/GregorR/rnnoise-models

@credits https://superuser.com/questions/733061/reduce-background-noise-and-optimize-the-speech-from-an-audio-clip-using-ffmpeg

@dependency proc-ffmpeg-wrapper.js

**/
proc['ffmpeg-denoise'] = {	
	pipe:pipe=>{		
		const isPreview = api.isPreviewMode(); //TODO
		const method = attr['method'] || 'FFT';
		const rnnModel = attr['rnn-model'] || 'beguiling-drafter';
		const rnnMix = attr['rnn-mix']*1||1;
		const rnnThreads = attr['rnn-threads']*1||1;
		const nlmStrengh = attr['nlm-strength']*1 || 0.00001;
		const nlmPatch = attr['nlm-patch']*1 || 0.002;
		const nlmResearch = attr['nlm-research']*1 || 0.006;
		const nlmSmooth = attr['nlm-smooth']*1 || 11;
		const nlmOutput = (attr['nlm-output'] || 'o').charAt(0);
		if(data && data.length){			
			log.warn(`Denoise currently does not handle input annotations, only pipeline corpus`);
		}			
		for(let f of pipe.corpus){			
			let step = log.addStep(f);			
			if(!f.media.size()) {
				step.error(`No media file available in ${f.id}`);		
				continue;
			}			
			for(let pmf of f.media){
				if(api.isEstimateMode()){
					api.converter.progressTodo(100*1000);
					continue;
				}
				step.file(pmf,`Input = `);			
				if(api.isPreviewMode()){
					step.info(`Not denoising because preview mode.`)
					continue;
				}			
				let outputFile = pipe.temp(pmf);
				
				//let filter = `arnndn="${api.doc.atk.path}/scripts/dependencies/rnnoise/${attrModel}.rnnn"`
				
				let filter = `afftdn`;
				switch(method){
					case 'NLM':
						filter = `anlmdn=`;
						filter += `s=${nlmStrengh}:`;
						filter += `p=${nlmPatch}:`;
						filter += `r=${nlmResearch}:`;
						filter += `m=${nlmSmooth}:`;
						filter += `o=${nlmOutput}`;
						break;
						
					case 'RNN':	
						let path = `${api.doc.atk.path}/scripts/dependencies/rnnoise/`;
						path = path.split(':').join('\\\\:');
						filter = `arnndn=m=${path}/${rnnModel}.rnnn:mix=${rnnMix}:threads=${rnnThreads}`;
				}
				let ffargs = [
					'-i',pmf.getAbsolutePath(),
					'-af',`"${filter}"`,
					outputFile.getAbsolutePath()
				]						
				//step.warn(ffargs.join(' '));
				step.info(`Filter: ${filter}`)
				step.file(outputFile,`Output = `)
				step.info(`Running denoise...`)
				step.work();
				
				let rs = ffmpegWrapper(step,ffargs,{
					onProgress:o=>api.converter.progressDone(o.percentChunk*1000),
				});
				if(!rs) return false;
				pmf.replaceWith(outputFile);			
			} // end of media
		} // end of corpus file
	}
}
