/** 
@author CG
@log 2024-02-19 initial implementation
@cachable

@attr include-tier-names bool (false) Whether to include the tier names in the subtitles
@attr include-tier-separator string (: ) A separator to add between tier name and subtitle text
@attr extend-duration-before int (0) Number of milliseconds to display subtitle before its original start time, so it is shown earlier
@attr extend-duration-after int (0) Number of milliseconds added after the original end time, so it stays visible longer
@attr style-color color (#FFFFFF) Subtitles color
@attr style-opacity string (100%) Subtitles opacity, in %
@attr style-outline-color color (#000000) Subtitles text outline color
@attr style-outline-opacity string (100%) Opacity of text outline, in %
@attr style-outline-width string () Width of the text outline, in pixels
@attr style-size int () Subtitles size
@attr style-bold bool (false) Subtitles weight
@attr style-font string () Subtitles font name
@param source-tier="" name="" | Add a tier from which to take subtitles text, and optionally customize its name
@dependency proc-ffmpeg-wrapper.js
@dependency milliseconds-to-time.js

Hardcodes annotations as subtitles on top of video. This processor will automatically use the values of the annotations that generated the clips, whenever they are available. It is possible to use different annotations from other tiers in range, by adding "source-tier" parameters.

**/
proc['hardsub'] = {	
	pipe:pipe=>{			
		const sourceTiers = {};
		const extendDurationBefore = attr['extend-duration-before']*1||0;
		const extendDurationAfter = attr['extend-duration-after']*1||0;
		const includeTierNames = attr['include-tier-names']=='true';
		const includeTierSeparator = attr['include-tier-separator']||': ';
		const styleAlpha = attr['style-opacity']||'100';
		const styleColor = attr['style-color']||'';
		const styleOutlineAlpha = attr['style-outline-opacity']||'100';
		const styleOutlineColor = attr['style-outline-color']||'';		
		const styleOutlineWidth = attr['style-outline-width']||'';
		const styleSize = attr['style-size']||'';
		const styleFont = attr['style-font']||'';
		const styleBold = attr['style-bold']=='true';
		
		const buildColor = (color,opacity)=>{
			let c = color || 'FFFFFF';
			if(c[0]=='#') c = c.substr(1);								
			if(c.length!=6 && c.length!=8) {
				log.warn(`invalid hexadecimal color value: ${color}`);
				c = 'FFFFFF';
			}
			if(opacity){
				if(c.length==6){
					let alpha = Math.round(opacity.replace('%','')*1);
					if(isNaN(alpha)){
						log.warn(`invalid alpha: ${opacity}`);
					}else{
						alpha = (255-Math.round((alpha/100)*255)).toString(16).toUpperCase();
						if(alpha.length===1) alpha = '0'+alpha;
						c = alpha + c;
					}
				}else{
					log.warn(`opacity is specified but color already includes alpha channel`);
				}
			}
			if(c.length==6) c = '00' + c;
			return c[0]+c[1] + // alpha
				c[6]+c[7] + // blue
				c[4]+c[5] + // green
				c[2]+c[3] ; // red			
		}
		
		const buildStyle = ()=>{			
			let styles = [];//
			if(styleColor || styleAlpha!='100') {								
				styles.push(`PrimaryColour=&H${buildColor(styleColor,styleAlpha)}&`);
			}
			if(styleOutlineColor || styleOutlineAlpha!='100') {								
				styles.push(`OutlineColour=&H${buildColor(styleOutlineColor,styleOutlineAlpha)}&`);
			}
			if(styleOutlineWidth){
				if(isNaN(styleOutlineWidth*1)) log.warn(`invalid style-outline-width: ${styleOutlineWidth}`);
				else styles.push(`Outline=${styleOutlineWidth}`);
			}
			if(styleSize){
				if(isNaN(styleSize*1)) log.warn(`invalid style-size: ${styleSize}`);
				else styles.push(`Fontsize=${styleSize}`);
			}
			if(styleFont){				
				styles.push(`Fontname=${styleFont}`);
			}
			if(styleBold){				
				styles.push(`Bold=1`);
			}
			//TODO styles.push(`BorderStyle=3,BackColour=&H${'CCFF00FF'}&`);
			return `:force_style='${styles.join(',')}'`;		
		}
		
		const forceStyle = buildStyle();
		
		for(let opt of params){
			if(opt.hasAttribute('source-tier')){
				let tier = opt.getAttribute('source-tier');
				let tierName = opt.hasAttribute('name') ? opt.getAttribute('name') : '';
				sourceTiers[tier] = tierName || tier;
			}
		}	
		for(let f of pipe.corpus){
			let step = log.addStep({name:f.id});
			if(!f.media.size()) {
				step.error(`No media file available in ${f.id}`);		
				continue;
			}			
			for(let pmf of f.media){
				if(pmf.isImage()) continue;
			    if(!pmf.exists()){
			        step.error(`The file does not exist: ${pmf.getAbsolutePath()}`);
			        continue;
			    }
				if(api.isEstimateMode()){		
					//step.info("ESTIMATE MODE DURATION "+pmf.mediaFile.getDuration())
					api.converter.progressTodo(pmf.mediaFile.getDuration());		
					continue;
				}
				let ma = pmf.modifiableAsset;
				if(!ma){
					step.warn(`Not a modifiable asset`);
					continue;
				}
				
				step.file(pmf,`Input = `);	
				
				let sourceAnnotation = ma.annotation;
				if(!sourceAnnotation){
					step.warn(`No annotation linked to this modifiable asset!`);
					continue;
				}
				
				if(api.isPreviewMode()){
					step.info(`Not applying because preview mode.`)
					continue;
				}			
				let outputFile = pipe.temp(pmf);
				let subFile = pipe.temp("srt");								
				let annotations = [];
				if(Object.keys(sourceTiers).length){					
					for(let tier in sourceTiers){
						let tt = sourceAnnotation.tier.af.getTier(tier);
						if(!tt) {
							step.warn(`No tier "${tt}" in "${sourceAnnotation.tier.af.id}"`);
							continue;
						}
						for(let a of tt.annotations){
							if(a.start >= sourceAnnotation.start
							&& a.stop <= sourceAnnotation.stop){
								//step.info(`Annotation in range: "${a.toString()}"`)
								annotations.push(a);
							}
						}
					}		
					const asort = (a,b)=>a.start>b.start?-1:1;
					annotations.sort(asort);
				}else{
					annotations.push(sourceAnnotation);
				}				
				if(!annotations.length){
					step.warn(`No annotation found in range of "${sourceAnnotation.toString()}"`);
					continue;					
				}
				let i = 0;
				let srt = '';				
				let shift = 0;
				for(let a of annotations){
					let start = a.start-sourceAnnotation.start-extendDurationBefore;
					let stop = a.stop-sourceAnnotation.start+extendDurationAfter;
					
					if(start<0) start = 0;
					
					let txt = a.value;
					
					if(includeTierNames){
						txt = (sourceTiers[a.tier.id] || a.tier.id) + includeTierSeparator + txt;
					}
					srt+=`${++i}\n${formatTimeSRT(start)} --> ${formatTimeSRT(stop)}\n${txt}\n\n`;
				}
				api.writeFile(subFile.getAbsolutePath(),srt);
				//step.file(subFile,`Built SRT`);
				step.file(outputFile,`Output = `)
				step.info(`Hardsubbing ${annotations.length} annotations`);
				step.work();
				let subFilePath = api.cleanPath(subFile.getAbsolutePath()).split(':').join('\\:')
				let ffargs = [
					'-i',pmf.getAbsolutePath(),							
					'-vf',`"subtitles='${subFilePath}'${forceStyle}"`,	
					'-c:a','copy',
					outputFile.getAbsolutePath()
				]		
				//step.info(ffargs.join(' '))
				let rs = ffmpegWrapper(step,ffargs,{					
					onProgress:o=>api.converter.progressDone(o.chunk),
					trap:s=>{
						if(s.substr(0,14)=='configuration:'){
							if(s.indexOf('--enable-libass')==-1){
								step.error(`ffmpeg is not built with libass enabled`);
								step.error(`please install a libass compatible ffmpeg version`);
								throw `ffmpeg is missing "libass" library`;
							}				
						}
					}
				});
				if(!rs) pipe.abort();
				pmf.replaceWith(outputFile);				
			}
		}		
	}
}
function formatTimeSRT(ms){
	let a = millisecondsToTime(ms).split('.');
	while(a[1].length<3) a[1] = a[1] + '0';
	return a.join(',');
}