/** 
@author CG
@log 2024-09-02 improved bridge
@log 2023-11-30 initial implementation
@experimental
@cachable

@attr file file () .R file to run
@attr source js () Plain R source code to run
@todo Add to source R functions to help interface with AVAA
@dependency proc-exec-output-wrapper.js
@dependency proc-r-utils.js

This processor executes a R program and integrates the resulting data into the final HTML page.
Resulting R output can be graphic files (jpg, png, gif, svg) or tabular text data.
Arguments provided to R are in order:
- temp directory path to work with and create result files
- path to a JSON file consisting of the selection (annotations) or data provided to the processor

R scripts must follow a specific input/output syntax to be compatible with AVAA (see "Calling R" in the scripting guide).


**/
proc['r-script'] = {	
	pipe:pipe=>{		
		if(api.isEstimateMode()) return true;
		const isPreview = api.isPreviewMode();
		const attrFile = attr['file'];		
		const attrSource = attr['source'];
		if(!attrFile && !attrSource){
			log.error(`you must specify the file (or source) attribute`);
			return false;
		}
		let rFile;
		if(attrSource){
			rFile = pipe.temp("r")
			api.writeFile(rFile,rHeader()+attrSource+rFooter());
		}else{
			rFile = api.getFile(attrFile);
			if(!rFile.exists()){
				log.error(`file not found "${attrFile}"`)
				return false;
			}
		}			
		let inputData = dataJSON;//JSON.stringify(data);
		try {
			if(JSON.parse(inputData)===null) inputData='{}';
		}catch(e){
			inputData='{}'
		}
		let inputFile = pipe.temp('json');
		api.writeFile(inputFile,inputData);
		
		const step = log.addStep({name:'R'})
		step.file(inputFile);
		//step.info(dataJSON)
		if(!RCheck(step)) return;
				//step.file(pmf,`Input = `);
				let tempdir = api.atk.getTempPath();
				let xargs = ['Rscript',
					`"${rFile.getAbsolutePath()}"`,
					`"${tempdir}"`,
					`"${api.cleanPath(inputFile.getAbsolutePath())}"`
				];
				//step.info(xargs.join(' '))
				//if(isPreview) xargs.push('--fake')				
				let aw = api.exec(xargs);
				step.monitor(aw);
				let progress = '';
				aw.onOutput(s=>{
					step.debug(s);
				})
				aw.onError(s=>{		
					if(!s) return;
					step.error(`Process Error: ${s}`)
				})
				aw.onOutput(execOutputWrapper(step,result=>{
					step.info(`Copying result to assets...`);
					//step.info(result);										
					let src = result.trim();
					let srcFile = new java.io.File(src);
					let target = "assets/" + srcFile.getName();										
					let targetFile = new java.io.File(api.doc.atk.path + "/" + target);
					let ext = org.apache.commons.io.FilenameUtils.getExtension(src).toLowerCase();
					api.doc.dependencies.put(target,"asset");
					org.apache.commons.io.FileUtils.copyFile(srcFile,targetFile);
					api.doc.html.append(`<img src="${api.toBaseURL(target)}"/>\n`)
					step.ok();					
					
				}));						
				let bench = Date.now();
				let exitStatus = aw.run();			
				let benched = Math.ceil((Date.now()-bench)/1000)+1;
				if(!exitStatus){
					step.info(`Completed in ${benched} seconds`);
					step.ok();
				}
				else {
					step.error(`Process exited with error code ${exitStatus} after ${benched} seconds`);
					return false;
				}				
				//continue;	
			//} // end of media
 		//} // end of domain file
		
		return true;
	}	
}

function rHeader(){
	return `#avaa generated r script ${Date.now()}
if("jsonlite" %in% rownames(installed.packages()) == FALSE) {install.packages("jsonlite")}
if("caTools" %in% rownames(installed.packages()) == FALSE) {install.packages("caTools")}
library(caTools)
library(jsonlite)
args <- commandArgs(trailingOnly = TRUE)
avaaTempDir <- args[1]
avaaInput <- fromJSON(txt=args[2])
avaaTemp <- function(ext){
	return(tempfile(pattern = "r-output-", fileext = paste(".",ext,sep=""), tmpdir = avaaTempDir))
}
avaaReturn <- function(output){
	cat(paste("#R:",output,"\n", sep=""))
}
avaaLogFile <- function(f,s=''){
	cat(paste("#F:",s," =", f, "\n", sep=""))	
}

`;
}
function rFooter(){
	return ``;
}
