/** 
@author CG
@log 2024-03-21 initial implementation

@input Array<Annotation>

@attr group regexp () Which group of corpus files to keep (regular expression)
@attr file regexp () Which file from the corpus to keep (regular expression)
@attr tag regexp () Files from the corpus to keep that have a tag satisfying this regular expression
@attr filter js () A custom filter function called for each file that should return true to keep the file, like (f)=>f.filename.includes('.eaf')

Reduces the corpus to specific files.
Useful to work on a subset of the corpus without modifying the corpus itself.
This processor also accepts a selection of annotations, in which case only corpus files of these annotations will be kept.

**/
PROC = {
	pipe:pipe=>{		
		if(!attr.group && !attr.file && !attr.filter && !attr.tag && !annotations.length){
			log.warn(`Please specify a criteria or make a selection to reduce the corpus!`)
			return;
		}
		
		const filterFunc = attr.filter ? new Function('return '+attr.filter)() : null;		
		const step = log.addStep({name:`Reducing Corpus`});		
		const corpusFilesToKeep = [];
		
		for(let f of pipe.corpus){			
			let keep=false;
			if(attr.group && f.af.group.match(attr.group)) keep = true;
			if(attr.file && f.af.filename.match(attr.file)) keep = true;
			if(attr.tag){
				for(let tag of f.af.tags) if(tag.match(attr.tag)) keep=true;
			}
			if(filterFunc && filterFunc(f.af)) keep = true;
			if(!keep) continue;			
			corpusFilesToKeep.push(f);			
		}	
		
		// now let's handle the selection of annotations, if any
		const afList = [];
		for(let a of annotations) if(afList.indexOf(a.af)<0) afList.push(a.af);

		// here it's tricky because we have a list of AnnotationFile
		// but the pipeline expects PipeDomainFile
		for(let af of afList){
			let pdf = pipe.getCorpusFile(af);
			if(!pdf){
				log.warn(`Failed to map corpus file ${af.id} to a pipeline corpus file`);
				continue;
			}
			if(corpusFilesToKeep.indexOf(pdf)<0) corpusFilesToKeep.push(pdf);
		}		
		
		// now we remove all corpus files from the pipeline, and add our selection
		pipe.corpus.clear();
		for(let f of corpusFilesToKeep){
			pipe.corpus.add(f);	
			step.info(`Keeping ${f.af.id}`);
		}
		
		if(!pipe.corpus.size()){
			log.warn(`The corpus is now empty, all files got excluded!`);
		}
		
	}
}
