/** 
@author CG
@log 2024-08-13 convert to wav if needed
@log 2024-08-07 use temporary json file to avoid cli limit
@log 2023-11-28 initial implementation
@license @ext Speaker diarization https://github.com/pyannote/pyannote-audio | MIT
@input Array<Annotation>
@output Array<Annotation>
@cachable
@attr @masked hf-token string () The HuggingFace access token, required to download pyannote models
@attr speakers int (0) Number of speakers in the audio

@todo usable without providing annotations

Speaker diarization is the process of marking segments of voice with their speaker. 
This processor takes a selection of annotations, and add to the corpus new annotations associated with their speaker tier.

@dependency proc-exec-output-wrapper.js
@dependency proc-python-wrapper.js
@dependency human-file-size.js

**/
proc['speaker-diarization-pyannote'] = {	
	data:data=>{		
		const isPreview = api.isPreviewMode();
		if(api.isEstimateMode()) return; //todo
		
		const hfToken = attr['hf-token'];				
		const pyannoteSpeakers = attr['speakers']*1 || 0;
				
		let req = log.addStep({name:'Requirements'})
						
		if(!hfToken){
			req.error("You must provide a HuggingFace token (hf-token attribute)")
			req.info("HuggingFace is a collaborative platform easing research in AI")
			req.info("The pyannote models are stored on HuggingFace and a token is required to access them")
			req.info("1) Create an account on https://huggingface.co")
			req.info("2) Accept conditions on https://huggingface.co/pyannote/segmentation-3.0")
			req.info("3) Accept conditions on https://huggingface.co/pyannote/speaker-diarization-3.1")
			req.info("4) Get your token here https://huggingface.co/settings/tokens")
			return;
		}
		
		if(!pythonCheck(req)) return;
		if(!pipCheck("torch",req)) return;
		if(api.isWindows()){
			if(!pipCheck("soundfile",req)) return;	
		}else{
			if(!pipCheck("sox",req)) return;	
		}
		if(!pipCheck("pyannote.audio",req)) return;
		if(!pipCheck("pyannote.core",req)) return;
		if(!pipCheck("onnxruntime",req)) return;		
		let awf = pythonWrapper("avaa-pyannote",req);
		if(!awf) return;
		
		if(!pyannoteSpeakers) log.warn(`For better results the "speakers" attributes should be set.`);			
		
		if(!annotations.size()){
			req.warn(`The processor expects a selection of annotations, for pyannote segments`);
			return;
		}
		
		const createdAnnotations = [];
		
		for(let f of pipe.corpus){

			let step = log.addStep(f);								
			// to pyannote we only need to provide segments
			let pyannoteAnnotations = annotations.toArray().filter(a=>a.af==f.af);
			
			if(!pyannoteAnnotations.length){
				step.info(`No annotation found in the selection for this file`);
				continue;
			}		
			
			step.info(`Found ${pyannoteAnnotations.length} annotations to diarize`);
			
			if(!f.media.size()) {
				step.error(`No media file available in ${f.id}`);		
				continue;
			}			
			for(let pmf of f.media){								
				if(!pmf.exists()){
					step.error(`The file does not exist: ${pmf.getAbsolutePath()}`);		
					continue;
				}				
				if(isPreview){
					step.info(`All good for preview mode`);
					continue;
				}
				step.file(pmf,`File = `);
				
				if(pmf.extension!='wav'){
					let wavFile = pipe.temp(pmf,"wav");
					step.file(wavFile, `Converting to WAV file `)
					step.work();							
					let rx = api.exec(['ffmpeg','-i',pmf.getAbsolutePath(),'-ar','44100',wavFile.getAbsolutePath()]);
					step.monitor(rx);				
					rx = rx.run();
					if(rx){
						step.error(`Something went wrong converting to WAV file.`);
						return false;// abort all
					}
					pmf = wavFile;
				}
				
				
				
				let annotationsFile = pipe.temp('json');
				let segments = pyannoteAnnotations.map(a=>({
					start:a.start,
					stop:a.stop
				}));
				api.writeFile(annotationsFile, JSON.stringify(segments));
				let xargs = ['python','-u',awf.getAbsolutePath(),
					'-i',pmf.getAbsolutePath(),
					'-t',hfToken,
					'-s',pyannoteSpeakers,					
					'-a',`"${annotationsFile.getAbsolutePath()}"` //btoa(JSON.stringify(pyannoteAnnotations))
				];
				//step.info(xargs.join(' '));
				//if(isPreview) xargs.push('--fake')				
				let aw = api.exec(xargs);
				step.monitor(aw);
				let progress = '';
				aw.onError(s=>{		
					if(!s) return;
					if(s.indexOf('The torchaudio backend is switched')>-1){
						//step.debug(s);
					}
					else if(s.indexOf('torchvision is not available')>-1){
						//step.debug(s);
					}
					else if(s.indexOf('UserWarning:')>-1) step.warn(s.split('UserWarning:')[1]);
					else if(s.trim().substring(0,16)=='warnings.warn(f"') return;
					else {
						if(s.substring(s.length-3)=='/s]') return; // process killed during downloading model
						step.error(`Process Error: ${s}`)
					}
				})
				
				aw.onOutput(execOutputWrapper(step,result=>{
					step.info(`Got results from diarization:`);
					step.info(result.substring(0,70)+' (...)');
					step.info(`Parsing diarization results...`);
					let a = JSON.parse(result);
					step.ok();					
					step.info(`Diarized ${a.length} segments`);					
					step.info(`Creating ${a.length} annotations`);		
					
					var i=0;
					for(o of a) createdAnnotations.push(f.addAnnotation({
						id:`pyannote-${o.start}-${o.stop}`,
						tier:(o.tier||'undefined').replace('SPEAKER_','speaker-'),
						start:o.start,
						stop:o.stop,
						value:pyannoteAnnotations[i++].value
					}));
					
					step.ok();
					
					
				},{
					progress:p=>{
						let a = p.split('/');
						let c = a[0]*1, t = a[1]*1;
						return `${a[2] ? 'Step - '+a[2] : 'Processing'} ${Math.floor((c/t)*100)}% ${c}/${t} ...`
					},
					download:p=>{
						let a = p.split('/');
						let c = a[0]*1, t = a[1]*1;
						return `Downloading model ${Math.floor((c/t)*100)}% ${humanFileSize(c)} / ${humanFileSize(t)} ...`;
					}
				}));						
				let bench = Date.now();
				let exitStatus = aw.run();			
				let benched = Math.ceil((Date.now()-bench)/1000)+1;
				if(!exitStatus){
					step.info(`Completed in ${benched} seconds`);
					step.ok();
				}
				else {
					step.error(`Process exited with error code ${exitStatus} after ${benched} seconds`);
					return false;
				}				
			} // end of media	
		}
		return createdAnnotations;	
	}
}

