/** 
@author CG
@experimental
@license @ext Noise-Robust Automatic Speech Recognizers are Also Strong Audio Event Taggers https://github.com/YuanGongND/whisper-at | BSD-2
@log 2023-11-28 initial implementation
@cachable
@output Array<Annotation>
@attr language string () The language to transcribe will also affect the names of the audio events
@attr model string (tiny) [tiny,small,medium,large-v3] The trained model to use for transcription
@todo @attr temperature number (0) Temperature, adjust to fix hallucinations
@todo @attr output-tier string (stt-whisper) Tier id for the extracted annotations
@todo @attr verbose bool (false) Whether to log the transcriptions as soon as they are detected

A variation of OpenAI Whisper designed to extract audio events of the 527-class AudioSet, Whisper-AT processor outputs general audio events as annotations.

@dependency proc-exec-output-wrapper.js
@dependency proc-python-wrapper.js
@dependency human-file-size.js

**/
proc['speech-to-text-whisper-at'] = {	
	pipe:pipe=>{		
		const isPreview = api.converter.getSetting('test-mode')=='true'; //todo api?
		if(api.converter.getSetting('estimate-mode')=='true') return; //todo api?
		const outputTier = attr['output-tier'] || 'stt-whisper';	
		
		const whisperModel = attr.model || 'small';
		const whisperTemp = attr.temperature || '0';
		const whisperLang = attr.language || '';
		//pipe.enter(log);
		let req = log.addStep({name:'Requirements'})
		let v = pythonCheck(req);
		if(!v) return;
		v = v.split('.').map(i=>i*1);
		/*if(v[0]>3){
			req.warn("Whisper will probably not work with Python version > 3")
		}
		if(v[0]>2 && v[1]>17){
			req.warn("Whisper will probably not work with Python version >= 3.18")
		}
		if(v[0]<3){
			req.warn("Whisper will probably not work with Python version < 3")
		}*/
		if(!pipCheck("numba",req)) return;
		if(!pipCheck("numpy",req)) return;
		if(!pipCheck("tqdm",req)) return;
		//if(!pipCheck("more-itert",req)) return;
		if(!pipCheck("torch",req)) return;
		if(api.isWindows()){
			if(!pipCheck("soundfile",req)) return;	
		}else{
			if(!pipCheck("sox",req)) return;	
		}
		if(!pipCheck("whisper-at",req)) return;
		let awf = pythonWrapper("avaa-whisper-at",req);
		if(!awf) return;
		
		const createdAnnotations = [];
		for(let f of pipe.corpus){			
			let step = log.addStep(f);								
			if(!f.media.size()) {
				step.error(`No media file available in ${f.id}`);		
				continue;
			}			
			for(let pmf of f.media){								
				if(!pmf.exists()){
					step.error(`The file does not exist: ${pmf.getAbsolutePath()}`);		
					continue;
				}				
				if(isPreview) continue;
				step.file(pmf,`Input = `);
				let xargs = ['python','-u',awf.getAbsolutePath(),
					'-i',pmf.getAbsolutePath(),
					'-m',whisperModel,
					'-t',whisperTemp,
					'-l',whisperLang
				];
				if(isPreview) xargs.push('--fake');
				let aw = api.exec(xargs);
				step.monitor(aw);
				let progress = '';
				aw.onError(s=>{		
					if(!s) return;
					if(s.indexOf('UserWarning:')>-1) step.warn(s.split('UserWarning:')[1]);
					else if(s.trim().substring(0,16)=='warnings.warn(f"') return;
					else {
						if(s.substring(s.length-3)=='/s]') return; // process killed during downloading model
						step.error(`Process Error: ${s}`)
					}
				})
				aw.onOutput(execOutputWrapper(step,result=>{
					step.info(`Got results from transcription:`);
					step.info(result.substring(0,70)+' (...)');
					step.info(`Parsing transcription results...`);
					let a = JSON.parse(result);
					step.ok();					
					step.info(`Transcribed ${a.length} segments`);					
					step.info(`Creating ${a.length} annotations`);		
					
					for(o of a) createdAnnotations.push(f.addAnnotation({
						id:`STT-${o.start}-${o.end}`,
						tier:outputTier,
						start:o.start*1000,
						stop:o.end*1000,
						value:(o.text+'').trim()
					}));
					
					step.ok();
					
					
				},{
					progress:p=>{
						let a = p.split('/');
						let c = a[0]*1, t = a[1]*1;
						return `Transcribing ${Math.floor((c/t)*100)}% ${c}/${t} ...`
					},
					download:p=>{
						let a = p.split('/');
						let c = a[0]*1, t = a[1]*1;
						return `Downloading model ${Math.floor((c/t)*100)}% ${humanFileSize(c)} / ${humanFileSize(t)} ...`;
					}
				}));				
				let exitStatus = aw.run();			
				let benched = Math.ceil(aw.duration/1000)+1;
				if(!exitStatus){
					step.info(`Completed in ${benched} seconds`);
					step.ok();
				}
				else {
					step.error(`Process exited with error code ${exitStatus} after ${benched} seconds`);
					return false;
				}				
			} // end of media	
		}
		return createdAnnotations;	
	}	
	
}
