/** 
@author CG
@experimental
@hidden
@license @ext An Optimized Speech-to-Text Pipeline for the Whisper Model Supporting Multiple Inference Engine https://github.com/shashikg/WhisperS2T | MIT
@log 2024-08-13 initial implementation
@cachable
@output Array<Annotation>
@attr language string () The language to transcribe, if not specified autodetect will be attempted
@attr model select (tiny) [tiny,small,medium,large,distil-large-v3] The trained model to use for transcription
@attr temperature number (0) Temperature, adjust to fix hallucinations
@attr device select (auto) [auto,cpu,cuda] The processing device to use
@attr precision select (int8) [auto,int8,fp16,fp32]
@attr beam-size int (5) The decoding beam size
@attr output-tier string (stt-whisper) Tier id for the extracted annotations
@attr verbose bool (false) Whether to log the transcriptions as soon as they are detected

A speech to text processor using https://github.com/shashikg/WhisperS2T to transcribe and automatically create annotations

@dependency proc-exec-output-wrapper.js
@dependency proc-python-wrapper.js
@dependency human-file-size.js

**/
proc['speech-to-text-whisper-s2t'] = {	
	pipe:pipe=>{		
		const isPreview = api.converter.getSetting('test-mode')=='true'; //todo api?
		if(api.converter.getSetting('estimate-mode')=='true') return; //todo api?
		const outputTier = attr['output-tier'] || 'stt-whisper';
		const whisperBeamSize = attr['beam-size']*1||5;
		const whisperPrecision = attr.precision || 'auto';
		const whisperDevice = attr.device || 'auto';
		const whisperModel = attr.model || 'small';
		const whisperTemp = attr.temperature || '0';
		const whisperLang = attr.language || '';
		//pipe.enter(log);
		let req = log.addStep({name:'Requirements'})
		let v = pythonCheck(req);
		if(!v) return;
		v = v.split('.').map(i=>i*1);
		if(v[0]>3){
			req.warn("Whisper will probably not work with Python version > 3")
		}
		if(v[0]>2 && v[1]>17){
			req.warn("Whisper will probably not work with Python version >= 3.18")
		}
		if(v[0]<3){
			req.warn("Whisper will probably not work with Python version < 3")
		}
		if(!pipCheck("torch",req)) return;
		if(api.isWindows()){
			if(!pipCheck("soundfile",req)) return;	
		}else{
			if(!pipCheck("sox",req)) return;	
		}
		if(!pipCheck("whisper-s2t",req)) return;
		let awf = pythonWrapper("avaa-whisper-s2t",req);
		if(!awf) return;
		
		const createdAnnotations = [];
		for(let f of pipe.corpus){			
			let step = log.addStep(f);								
			if(!f.media.size()) {
				step.error(`No media file available in ${f.id}`);		
				continue;
			}			
			for(let pmf of f.media){								
				if(!pmf.exists()){
					step.error(`The file does not exist: ${pmf.getAbsolutePath()}`);		
					continue;
				}				
				if(isPreview) continue;
				step.file(pmf,`Input = `);
				let xargs = ['python','-u',awf.getAbsolutePath(),
					'-i',pmf.getAbsolutePath(),
					'-m',whisperModel,
					'-t',whisperTemp,
					'-l',whisperLang,
					'-b',whisperBeamSize,
					'-d',whisperDevice,
					'-p',whisperPrecision
					
				];
				if(isPreview) xargs.push('--fake');
				let aw = api.exec(xargs);
				step.monitor(aw);
				let progress = '';
				aw.onError(s=>{		
					if(!s) return;
					if(s.indexOf('symlinks')>-1) return;
					if(s.indexOf('UserWarning:')>-1) step.warn(s.split('UserWarning:')[1]);
					else if(s.trim().substring(0,15)=='warnings.warn(m') return;
					else {
						if(s.substring(s.length-3)=='/s]') return; // process killed during downloading model
						step.error(`Process Error: ${s}`)
					}
				})
				aw.onOutput(execOutputWrapper(step,result=>{
					step.info(`Got results from transcription:`);
					step.info(result.substring(0,70)+' (...)');
					//return;
					step.info(`Parsing transcription results...`);
					let a = JSON.parse(result);
					step.ok();					
					step.info(`Transcribed ${a.length} segments`);					
					step.info(`Creating ${a.length} annotations`);		
					
					for(o of a){
						let v = fixWhisperValue(o.text,step);
						if(v===null) continue;
						createdAnnotations.push(f.addAnnotation({
							id:`STT-${o.start}-${o.end}`,
							tier:outputTier,
							start:o.start*1000,
							stop:o.end*1000,
							value:v
						}));
					}
					step.ok();
					
					
				},{
					progress:p=>{
						let a = p.split('/');
						let c=a[0]*1, t=a[1]*1, C=a[2]*1||0;
						return `Transcribing ${Math.floor((c/t)*100)}% ${Math.round(c)}/${Math.round(t)} (${C} annotations transcribed) ...`
					},
					download:p=>{
						let a = p.split('/');
						let c = a[0]*1, t = a[1]*1;
						return `Downloading model ${Math.floor((c/t)*100)}% ${humanFileSize(c)} / ${humanFileSize(t)} ...`;
					}
				}));				
				let exitStatus = aw.run();			
				let benched = Math.ceil(aw.duration/1000)+1;
				if(!exitStatus){
					step.info(`Completed in ${benched} seconds`);
					step.ok();
				}
				else {
					step.error(`Process exited with error code ${exitStatus} after ${benched} seconds`);
					return false;
				}				
			} // end of media	
		}
		return createdAnnotations;	
	}	
	
}

function fixWhisperValue(v,step){
	//TODO language aware
	v = (v || '').trim();
	for(let a of Object.values(WHISPER_BIAS)){
		for(let s of a){			
			if(s==v){
				step && step.warn('cleaned one "${v}"');
				return null;
			}
		}
	}
	return v;
}

//https://github.com/openai/whisper/discussions/928
const WHISPER_BIAS = {
  "en": [
    " www.mooji.org",
  ],
  "nl": [
    " Ondertitels ingediend door de Amara.org gemeenschap",
    " Ondertiteld door de Amara.org gemeenschap",
    " Ondertiteling door de Amara.org gemeenschap"
  ],
  "de": [
    " Untertitelung aufgrund der Amara.org-Community",
    " Untertitel im Auftrag des ZDF für funk, 2017",
    " Untertitel von Stephanie Geiges",
    " Untertitel der Amara.org-Community",
    " Untertitel im Auftrag des ZDF, 2017",
    " Untertitel im Auftrag des ZDF, 2020",
    " Untertitel im Auftrag des ZDF, 2018",
    " Untertitel im Auftrag des ZDF, 2021",
    " Untertitelung im Auftrag des ZDF, 2021",
    " Copyright WDR 2021",
    " Copyright WDR 2020",
    " Copyright WDR 2019",
    " SWR 2021",
    " SWR 2020",
  ],
  "fr": [
    " Sous-titres réalisés para la communauté d'Amara.org",
    " Sous-titres réalisés par la communauté d'Amara.org",
    " Sous-titres fait par Sous-titres par Amara.org",
    " Sous-titres réalisés par les SousTitres d'Amara.org",
    " Sous-titres par Amara.org",
    " Sous-titres par la communauté d'Amara.org",
    " Sous-titres réalisés pour la communauté d'Amara.org",
    " Sous-titres réalisés par la communauté de l'Amara.org",
    " Sous-Titres faits par la communauté d'Amara.org",
    " Sous-titres par l'Amara.org",
    " Sous-titres fait par la communauté d'Amara.org",
    " Sous-titrage ST' 501",
    " Sous-titrage ST'501",
    " Cliquez-vous sur les sous-titres et abonnez-vous à la chaîne d'Amara.org",
    " ❤️ par SousTitreur.com",
  ],
  "it": [
    " Sottotitoli creati dalla comunità Amara.org",
    " Sottotitoli di Sottotitoli di Amara.org",
    " Sottotitoli e revisione al canale di Amara.org",
    " Sottotitoli e revisione a cura di Amara.org",
    " Sottotitoli e revisione a cura di QTSS",
    " Sottotitoli e revisione a cura di QTSS.",
    " Sottotitoli a cura di QTSS",
  ],
  "es": [
    " Subtítulos realizados por la comunidad de Amara.org",
    " Subtitulado por la comunidad de Amara.org",
    " Subtítulos por la comunidad de Amara.org",
    " Subtítulos creados por la comunidad de Amara.org",
    " Subtítulos en español de Amara.org",
    " Subtítulos hechos por la comunidad de Amara.org",
    " Subtitulos por la comunidad de Amara.org",
    " Más información www.alimmenta.com",
    " www.mooji.org",
  ],
  "gl": [
    " Subtítulos realizados por la comunidad de Amara.org"
  ],
  "pt": [
    " Legendas pela comunidade Amara.org",
    " Legendas pela comunidade de Amara.org",
    " Legendas pela comunidade do Amara.org",
    " Legendas pela comunidade das Amara.org",
    " Transcrição e Legendas pela comunidade de Amara.org"
  ],
  "la": [
    " Sottotitoli creati dalla comunità Amara.org",
    " Sous-titres réalisés para la communauté d'Amara.org"
  ],
  "ln": [
    " Sous-titres réalisés para la communauté d'Amara.org"
  ],
  "pl": [
    " Napisy stworzone przez społeczność Amara.org",
    " Napisy wykonane przez społeczność Amara.org",
    " Zdjęcia i napisy stworzone przez społeczność Amara.org",
    " napisy stworzone przez społeczność Amara.org",
    " Tłumaczenie i napisy stworzone przez społeczność Amara.org",
    " Napisy stworzone przez społeczności Amara.org",
    " Tłumaczenie stworzone przez społeczność Amara.org",
    " Napisy robione przez społeczność Amara.org",
    " www.multi-moto.eu",
  ],
  "ru": [
    " Редактор субтитров А.Синецкая Корректор А.Егорова"
  ],
  "tr": [
    " Yorumlarınızıza abone olmayı unutmayın.",
  ],
  "su": [
    " Sottotitoli creati dalla comunità Amara.org"
  ],
  "zh": [
    "字幕由Amara.org社区提供",
    "小編字幕由Amara.org社區提供"
  ]
}