/** 
@author CG
@license @ext Pre-trained enterprise-grade Voice Activity Detector https://github.com/snakers4/silero-vad | MIT
@license Silero JIT and ONNX files | MIT
@log 2023-12-15 adapted for cache
@log 2023-11-29 initial implementation

@cachable
@output Array<Annotation>

@attr output-tier string (vad-silero) Tier id for the generated annotations
@attr threshold number (0.5) Use a higher threshold for noisy audio
@attr sampling-rate number (16000) [8000,16000,32000,48000]
@attr min-silence-duration int (500) Number of milliseconds
@attr min-speech-duration int (1000) Minimum duration (in milliseconds) of activity to consider a voice segment

@dependency proc-exec-output-wrapper.js
@dependency proc-python-wrapper.js

Silero's Voice Activity Detector processor creates annotations for each segment of input audio containing voice.

**/
proc['vad-silero'] = {	
	pipe:pipe=>{		
		const isPreview = api.isPreviewMode();
		if(api.isEstimateMode()) return; //todo
		const outputTier = attr['output-tier'] || 'vad-silero';		
		const outputMode = attr['output'] || 'domain';
		const vadThreshold = attr['threshold']*1 || 0.5;
		const vadSamplingRate = attr['sampling-rate']*1 || 16000;
		const vadSilence = attr['min-silence-duration']*1 || 500;
		const vadSpeech = attr['min-speech-duration']*1 || 1000;

		let req = log.addStep({name:'Requirements'})
		if(!pythonCheck(req)) return;
		if(!pipCheck("torch",req)) return;		
		
		if(api.isWindows()){
			if(!pipCheck("soundfile",req)) return;	
		}else{
			if(!pipCheck("sox",req)) return;	
		}
		
		if(!(wrapper=pythonWrapper("avaa-vad-silero",req))) return;	
		
		const createdAnnotations = [];
		for(let f of pipe.domain){			
			let step = log.addStep(f);						
			if(!f.media.size()) {
				step.error(`No media file available in ${f.id}`);		
				continue;
			}			
			for(let pmf of f.media){
				let media = pmf;
				//TODO estimatemode
				if(isPreview) continue;
				step.file(media,`Input =`);							
				if(media.extension!='wav'){
					let wavFile = pipe.temp(pmf,"wav");
					step.file(wavFile, `Converting to WAV file `)
					step.work();							
					let rx = api.exec(['ffmpeg','-i',pmf.getAbsolutePath(),'-ar','16000',wavFile.getAbsolutePath()]);
					step.monitor(rx);				
					rx = rx.run();
					if(rx){
						step.error(`Something went wrong converting to WAV file.`);
						return false;// abort all
					}
					media = wavFile;
					step.ok();					
				}						
				step.info(`Launching VAD process`)
				step.work();
				let xargs = ['python',
					'-u',wrapper.getAbsolutePath(),
					'-i',media.getAbsolutePath(),				
					'-t',vadThreshold,
					'-s',vadSilence,
					'-e',vadSpeech,
					'-r',vadSamplingRate
				];
				if(isPreview) xargs.push('--fake')			
				let aw = api.exec(xargs);
				let progress = '';
				aw.onError(s=>{
					step.error(`A process error occurred: ${s}`)
				})
				let results = [];
				aw.onOutput(execOutputWrapper(step,result=>{
					step.info(`Got results from detection:`);
					step.info(result.substring(0,70)+' (...)');
					step.info(`Parsing silero results`);
					results = JSON.parse(result);
					step.ok();
				},{progress:p=>`Processing... ${p}`}));
					
				let bench = Date.now();
				step.monitor(aw);
				let exitStatus = aw.run();			
				let benched = Math.ceil((Date.now()-bench)/1000)+1;
				if(!isPreview && !exitStatus && !results.length){
					step.warn(`No voice activity found.`)
				}	
				if(results.length){
					step.info(`Creating ${results.length} annotations`);
					for(let o of results){					
						createdAnnotations.push(f.addAnnotation({
							id:`VA-${o.start}-${o.end}`,
							tier:outputTier,
							start:o.start*1000,
							stop:o.end*1000,
							value:'VAD'
						}));
					}
					step.ok();
				}				
				if(!exitStatus){
					step.info(`Completed in ${benched} seconds`);
					step.ok();
				}else{
					step.error(`Process exited with error code ${exitStatus} after ${benched} seconds`);
					return false;
				}
				continue;
			} // end of media
		} // end of domain file
		//log.info(`Generated ${createdAnnotations.length} annotations`);
		return createdAnnotations;
	}
}