/** 
@author CG
@license @ext Video anonymization by face detection https://github.com/ORB-HD/deface | MIT
@license Video filters by https://github.com/dyne/frei0r | GPL-2.0
@license Frei0r DLL pack for Windows by Gyan Doshi https://www.gyan.dev/ffmpeg | CC BY-NC-ND 4.0
@log 2023-12-11 initial implementation
@log 2024-03-25 added deface options
@experimental
@cachable

@attr mode select () [,deface,cartoon,cartoon-blur,retro-glow] The type of anonymisation to apply on the video
@attr deface-mode select (blur) [blur,solid,mosaic] Anonymization filter mode for face regions
@attr deface-threshold number (0.2) Detection threshold (tune this to trade off between false positive and false negative rate)
@attr deface-mask-scale number (1.3) Scale factor for face masks, to make sure that masks cover the complete face
@attr deface-mosaicsize int (20) Width of the mosaic squares when deface-mode is mosaic
@attr deface-boxes bool (false) Use boxes instead of ellipse masks
@attr deface-draw-scores bool (false) Draw detection scores onto outputs, useful to find the best threshold
@attr deface-downscale string () Downscale resolution for the network inference (WxH)
@dependency proc-ffmpeg-wrapper.js
@dependency proc-exec-output-wrapper.js
@dependency proc-python-wrapper.js

Anonymize videos with these special effects:
	- deface: automatically detects and blur faces
	- cartoon: cartoonize the video
	- cartoon-blur: cartoonize and blur the video

**/
proc['video-anonymizer'] = {	
	pipe:function(pipe){		
		if(!attr.mode){
			log.error(`Please specify an anonymisation mode`);
			return false;			
		}
		const attrSteps = attr.steps == 'true';
		const attrMode = attr.mode.split('-').join('_');
		const modeFunc = this['mode_'+attrMode];
		
		if(!modeFunc){
			log.error(`Unknow anonymisation mode: "${attr.mode}"`);
			return false;
		}
				
		if(attrMode=='deface'){
			let req = log.addStep({name:'Requirements'})
			if(!pythonCheck(req)) return false;
			if(!pipCheck("deface",req)) return false;
		}
				
		for(let f of pipe.domain){
			let step = log.addStep({name:f.id});
			if(!f.media.size()) {
				step.error(`No media file available in ${f.id}`);		
				continue;
			}			
			for(let pmf of f.media){
			    if(!pmf.exists()){
			        step.error(`The file does not exist: ${pmf.getAbsolutePath()}`);
			        continue;
			    }
				if(api.isEstimateMode()){
					api.converter.progressTodo(pmf.getDuration());		
					continue;
				}
				//step.info(`Mode: ${attrMode}`);
				if(api.isPreviewMode()){
					step.info(`Not applying because preview mode.`)
					continue;
				}			
				let out = modeFunc(pipe,step,pmf);				
				if(out) pmf.replaceWith(out);
				else {
					log.error(`Something went wront during video anonymisation with mode "${attrMode}"`);
					pipe.abort();		
				}					
			}
		}		
	},	
	old_mode_cartoon:(pipe,step,pmf)=>{
		const FREI0R_PATH = api.atk.path+'/scripts/dependencies/frei0r';
		var steps = 0;
		// let's explore a chainable solution
		// seems not as efficient as concating the filters for only one ffmpeg call
		const frei0rize = (input,filter,params)=>{
			steps++;
			let outputFile = pipe.temp(pmf);
			step.file(input,`Step ${steps}) Input = `);	
			step.file(outputFile,`Step ${steps}) Output = `)
			let ffargs = [
				'-i',input.getAbsolutePath(),										
				'-vf',`"frei0r=filter_name=${filter}:${params.join(':')}"`,			//
				outputFile.getAbsolutePath()
			]		
			step.info(`frei0rizing with ${filter}=${params} ...`);
			step.work();	
			let rs = ffmpegWrapper(step,ffargs,{
				env:{FREI0R_PATH:FREI0R_PATH},
				onProgress:o=>api.converter.progressDone(o.chunk),
				trap:s=>{
					if(s.substr(0,14)=='configuration:'){
						if(s.indexOf('--enable-frei0r')==-1){
							step.error(`ffmpeg is not built with frei0r enabled`);
							step.error(`please install full version of ffmpeg`);
						}else{
							//step.info(`frei0r is enabled`);
						}					
					}
				}
			});
			if(!rs) throw `Failed applying the frei0r filter`;
			return outputFile;
		}
				
		let cartoonized = frei0rize(pmf,'cartoon',['0.9995']);
		//let blurred = frei0rize(cartoonized,'IIRblur',['0.4']);
		
		return cartoonized;
	},
	mode_cartoon:(pipe,step,pmf)=>{
		let filters = [
			frei0rize('cartoon',['0.9995']),			
			'format=yuv420p'
		]
		return runfrei0r(pipe,step,pmf,filters)		
	},	
	mode_cartoon_blur:(pipe,step,pmf)=>{
		let filters = [
			frei0rize('cartoon',['0.9995']),
			frei0rize('IIRblur',['0.4']),
			'format=yuv420p'
		]
		return runfrei0r(pipe,step,pmf,filters)		
	},
	mode_deface:(pipe,step,pmf)=>{
		const defaceAttributes = {};
		const defaceMap = {threshold:'thresh',mode:'replacewith',downscale:'scale'}
		for(let k in attr) if(k.includes('deface-')) {
			let dk = k.replace('deface-','');
			dk = defaceMap[dk] || dk;
			defaceAttributes[dk] = attr[k];
		}
		//step.info(JSON.stringify(defaceAttributes));		
		step.file(pmf,`Input = `);
		let outputFile = pipe.temp(pmf);		
		step.file(outputFile,`Output = `)
		step.info(`Defacing 0% ...`);
		step.work();
		let xargs = ['deface',pmf.getAbsolutePath(),'--output',outputFile.getAbsolutePath(),'--keep-audio'];
		for(let k in defaceAttributes){
			let v = defaceAttributes[k];
			xargs.push('--'+k);
			if(v==='true') continue;
			xargs.push(v)
		}
		//if(isPreview) xargs.push('--fake')			
		let proc = api.exec(xargs);
		let progress = '';
		proc.onError(s=>{
			if(!s || !s.trim()) return;
			if(s.indexOf('%')>-1){
				let perc = s.split('|')[0]
				step.replace(`Defacing ${perc} ...`);				
			}
			else step.error(`A process error occurred: ${s}`)
		})			
		proc.onOutput(s=>{
			s = s.trim();
			if(s.substring(0,6)=='Input:'){
				//step.file(s.substring(7).trim(),'Input = ');
			}
			if(s.indexOf('%')>-1){
				let perc = s.split('|')[0]
				step.replace(`Defacing ${perc} ...`);				
			}
			//else step.debug(s);
		});
		step.monitor(proc);			
		let exitStatus = proc.run();			
		let benched = Math.ceil(proc.duration/1000);
		if(!exitStatus){
			step.replace(`Defacing 100% ...`);
			step.ok();
			step.info(`Completed in ${benched} seconds`);
			step.ok();
		}else{
			step.error(`Process exited with error code ${exitStatus} after ${benched} seconds`);
			pipe.abort();
		}
		return outputFile;
	},
	mode_retro_glow:(pipe,step,pmf)=>{
		//https://www.reddit.com/r/ffmpeg/comments/im2mkp/creating_a_retro_glow_effect_with_ffmpeg/
		let outputFile = pipe.temp(pmf);
		step.file(pmf,`Input = `);	
		step.file(outputFile,`Output = `)		 
		let ffargs = [
			'-i',pmf.getAbsolutePath(),										
			'-filter_complex',`"[0]rgbashift=rh=15:bv=15:gh=-15[rgba-shifted];[rgba-shifted]format=rgba[in2];[in2][0]scale2ref[in2][in1];[in1][in2]blend=overlay[overlay];[overlay]gblur=sigma=42:steps=6[gblur];[gblur]format=yuv420p[in2];[in2][0]scale2ref[in2][in1];[in1][in2]blend=screen"`,
			'-pix_fmt','yuv420p',
			outputFile.getAbsolutePath()
		]		
		step.info(ffargs.join(' '));
		step.work();	
		let rs = ffmpegWrapper(step,ffargs,{			
			onProgress:o=>api.converter.progressDone(o.chunk)			
		});
		if(!rs) throw `Failed applying the video filters`;
		return outputFile;
	}
}


function runfrei0r(pipe,step,pmf,filters){
	const FREI0R_PATH = api.atk.path+'/scripts/dependencies/frei0r';		
	let outputFile = pipe.temp(pmf);
	step.file(pmf,`Input = `);	
	step.file(outputFile,`Output = `)
	let f = filters.join(',');
	if(api.isWindows()) f = `"${f}"`;
	let ffargs = [
		'-i',pmf.getAbsolutePath(),										
		'-vf',f,
		outputFile.getAbsolutePath()
	]		
	//step.debug(`frei0rizing with ${filters} ...`);
	step.work();	
	let rs = ffmpegWrapper(step,ffargs,{
		env:{FREI0R_PATH:FREI0R_PATH},
		onProgress:o=>api.converter.progressDone(o.chunk),
		trap:s=>{
			if(s.substr(0,14)=='configuration:'){
				if(s.indexOf('--enable-frei0r')==-1){
					step.error(`ffmpeg is not built with frei0r enabled`);
					step.error(`please install full version of ffmpeg`);
				}else{
					//step.info(`frei0r is enabled`);
				}					
			}
		}
	});
	if(!rs) throw `Failed applying the video filters`;
	return outputFile;
}
function frei0rize(filter,params){
	let f = `frei0r=filter_name=${filter}:filter_params=${params.join('|')}` 	
	return f;
}