/** 
@author CG
@log 2024-03-06 initial implementation
@label add an Advene AZP file
@hidden
@attr file file () The .AZP file to load
@attr media file () Optional media file, if it can't be found via AZP
@attr group string () The group to associate to this corpus file and all its annotations
@attr tags tags () A list of tags associated to this corpus file

Provides corpus mapping for Advene annotation format

@dependency human-file-size.js
@dependency proc-ffmpeg-wrapper.js
@register extension azp
**/
provider.advene = {
	load:function(){
		const step = log.addStep({name:`Loading AZP`});
		if(!attr.file){
			step.error(`No .AZP file was specified`);
			return;
		}
				
		const file = new java.io.File(attr.file);
		if(!file.exists()){
			step.error(`File "${attr.file}" does not exist`);
			return;
		}
		step.info(`Loading content...`);
		const zip = new java.util.zip.ZipFile(file);
		const entry = zip.getEntry("content.xml");
		const stream = zip.getInputStream(entry);
		const content = api.readStream(stream);	
		if(!content){
			step.error(`Could not read "content.xml"`);
			return;
		}  
		step.ok();

		// create a generic AnnotationFile instance
		let af = api.createAnnotationFile(file);
		
		if(attr.group) af.group = attr.group;
		if(attr.tags) af.tags.addAll(attr.tags.split(','));		
		
		api.doc.corpus.add(af);
		
		step.info(`Parsing content...`); 
		// Rhino supports E4X so why not use it
		try {
			const x = new XML(content);
		}catch(ex){
			step.error(`XML content parsing failed "${ex.message}"`);
			return false;
		}
		step.ok();
		
		const findAdveneMedia = (path)=>{
			let f = new java.io.File(path);
			if(!f.exists()) f = new java.io.File(file.getParent()+'/'+f.getName());// same folder maybe
			if(!f.exists()) return null;
			return f;
		}
		//step.info(`Creator: ` + x.@*::creator);		
		let mediaExport = x..*::['website-export-video-url']+'';
		let mediaURI = x..*::media_file+'';
		if(!mediaURI) mediaURI = x..*::media_uri+'';
		if(!mediaURI){			
			step.warn(`No media file linked`);
		}else if(mediaURI.substring(0,7)!='file://'){
			step.warn(`Can't understand mediaURI: "${mediaURI}"`)			
		}else{
			let warn404 = [];
			mediaURI = decodeURIComponent(mediaURI);
			let mediaFile = findAdveneMedia(mediaURI.substring(8));
			if(!mediaFile){ // lol if cascade refactor please
				warn404.push(`Could not find media @ ${mediaURI}`);
				if(mediaExport){
					mediaFile = findAdveneMedia(mediaExport);
					if(!mediaFile){
						warn404.push(`Could not find media @ ${mediaExport}`);
						if(attr.media){
							mediaFile = findAdveneMedia(attr.media);
							if(!mediaFile){
								warn404.push(`Could not find media @ ${attr.media}`);
							}
						}
					}
				}
			}
			if(mediaFile){
				af.addMediaFile(mediaFile);	
				step.file(mediaFile,`Found media file`);
			}else{
				warn404.forEach(s=>step.warn(s));
			}
		}
		//['']
		let tiers = {};
		for each(let t in x..*::['annotation-type']){
			let id = t..@*::id;			
			if(!tiers[id]) tiers[id] = af.addTier({id:id});
			else step.warn(`Tier declared twice: ${id}`); // heh who knows
		}		
		
		let annotations = {};
		for each(let a in x..*::annotation){
			let id = a..@*::id;
			let type = a..@*::type;
			let value = a..*::content+'';
			if(!type){
				step.warn(`Annotation id="${id}" has no type relation`);
				continue;
			}
			if(type.charAt(0)!='#'){
				step.warn(`Annotation id="${id}" type relation should start with #`);
				continue;
			}
			let tier = tiers[type.substring(1)];
			if(!tier){
				step.warn(`Annotation id="${id}" type="${type}" has no declared matching tier`);
				continue;
			}
			//step.info(`${id} : ${value}`)
			let start=a..@*::begin*1||0;
			let stop=a..@*::end*1||0;			
			annotations[id] = af.addAnnotation({
				tier:tier,
				id:id,
				start:start,
				stop:stop,
				value:value
			});			
		}		
		step.info(`Found ${Object.keys(tiers).length} tiers and ${Object.keys(annotations).length} annotations`);
		return af;
	}
}