/** 
@author CG
@log 2024-02-23 migrated to provider system
@label add a subtitled MKV file
@hidden
@attr file file () The .MKV file to load
@attr tier string (mkv) A tier name to link to the SRT annotations
@attr index int (0) The index of the subtitle track to extract
@attr group string () The group to associate to this corpus file and all its annotations
@attr tags tags () A list of tags associated to this corpus file
Provides domain mapping for the SubRip annotation format
https://en.wikipedia.org/wiki/SubRip

@dependency human-file-size.js
@dependency proc-ffmpeg-wrapper.js
@register extension mkv
**/

provider.matroska = {
	load:function(){
		const step = log.addStep({name:`Loading MKV`});
		const index = attr.index*1 || 0;
		if(!attr.file){
			step.error(`No .MKV file was specified`);
			return;
		}
				
		let file = new java.io.File(attr.file);
		if(!file.exists()){
			step.error(`File "${attr.file}" does not exist`);
			return;
		}
		
		let srtFile = new java.io.File(file.getParent() + "/"+file.getName()+`.${index}.srt`);
		if(srtFile.exists()){
			step.info(`Subtitles already extracted`);			
			step.ok();
		}else{
			let ffargs = ['-y',
				'-i',file.getAbsolutePath(),														
				'-map', `0:s:${index}`,
				srtFile.getAbsolutePath()
			]					
			step.info(`Extracting subtitles...`)
			step.work();						
			let rs = ffmpegWrapper(step,ffargs);
			if(!rs){
				step.error(`Something went wrong executing ffmpeg`);
				return false;
			}			
			if(!srtFile.exists()){
				step.error(`No subtitle track seems available in the MKV file`);
				return false;
			}
		}
		//step.file(srtFile, `Reading`);
		let content = api.readFile(srtFile);
		if(!content){
			step.error(`Could not read file ${output}`);
			return;
		}        
        step.ok();
		// create a generic AnnotationFile instance
		const af = api.createAnnotationFile(file);
		if(attr.group) af.group = attr.group;
		if(attr.tags) af.tags.addAll(attr.tags.split(','));
		
		api.doc.domain.add(af);
		
		// the tier that will hold the annotations
		let tier = af.addTier({id:attr.tier || 'mkv'});
        
		// parsing srt
		let lines = content.split("\n");
        for (let i = 0; i < lines.length; i++) lines[i] = lines[i].split("\r").join("");
        for (let i = 0; i < lines.length; i+=0) {
            let id = lines[i];
			let codeLine = lines[i + 1];
			if(!id && !codeLine) break;
			if(!codeLine){
				step.error(`Missing timecode line after id "${id}" line ${i}`);
				break;
			}
            let times = codeLine.substring(0,29);
            let codes = times.split("-->");
            if (codes.length != 2) {
				step.error(`Can't parse srt time codes: ${times}`);
				break;
			}
            let start = parseTimeCode(codes[0]);
            let stop = parseTimeCode(codes[1]);
            let value = "";
            let j = 0;
            while (true) {
                j++;
                let p = i + 1 + j;
                if (p >= lines.length) break;
                let lineOfText = lines[p];
                if (!lineOfText) break;
                value += !value ? lineOfText : "\n" + lineOfText;
            }
            i += 2 + j;
			// insert annotation into corpus file
			af.addAnnotation({
				tier:tier,
				id:id,
				start:start,
				stop:stop,
				value:value
			});
        }
		if(!af.annotations.size()){
			step.warn(`No annotation found in the file!`);
		}else{
			step.info(`Found ${af.annotations.size()} annotations`);
			step.ok();
		}	
				
		af.addMediaFile(file);		
		return af;
	}	
}


function parseTimeCode(code){
	code = code.trim();
	t = code.split(",");
	ms = t[1]*1;
	t = t[0].split(":");
	h = t[0]*1;
	m = t[1]*1;
	s = t[2]*1;
	return (h*3600+m*60+s) * 1000 + ms;
}