/** 
@author CG
@log 2024-04-01 initial implementation
@label add an oTranscribe OTR file
@hidden
@attr file file () The .OTR file to load
@attr tier-separator () A separator to extract tier from a transcript line
@attr media file () The media file associated to this oTranscribe file, if it can't be found in the same folder

@attr group string () The group to associate to this corpus file and all its annotations
@attr tags tags () A list of tags associated to this corpus file

Provides corpus mapping for oTranscribe annotation format
@image png
@register extension otr
**/
provider.otranscribe = {
	load:function(){
		const tierSeparator = attr['tier-separator'] || '';
		const step = log.addStep({name:`Loading oTranscribe`});
		if(!attr.file){
			step.error(`No .OTR file was specified`);
			return;
		}
				
		const file = new java.io.File(attr.file);
		if(!file.exists()){
			step.error(`File "${attr.file}" does not exist`);
			return;
		}
		step.info(`Loading content...`);
		const content = api.readFile(file);	
		if(!content){
			step.error(`Could not read "${attr.file}"`);
			return;
		}  
		step.ok();
		
		// create a generic AnnotationFile instance
		const af = api.createAnnotationFile(file);
		if(attr.group) af.group = attr.group;
		if(attr.tags) af.tags.addAll(attr.tags.split(','));
		api.doc.domain.add(af);
		
		step.info(`Parsing content...`); 		
		// oTranscribe is a JSON file so we parse it first
		try {
			const otr = JSON.parse(content);
		}catch(ex){
			step.error(`Could not parse "${attr.file}": ${ex.message}`);
			return;
		}
		if(otr==null){
			step.error(`Could not parse "${attr.file}", resulting JSON is null`);
			return;
		}
		if(!otr.text){
			step.error(`Could not find "text" property in parsed JSON`);
			return;
		}
		step.ok();
		const findOTRMedia = (path)=>{
			let f = new java.io.File(path);
			if(!f.exists()) f = new java.io.File(file.getParent()+'/'+f.getName());// same folder maybe
			if(!f.exists()) return null;
			return f;
		}
		let otrMedia;
		// let's find the media file
		if(otr.media){
			//step.info(`OTR media should be: ${otr.media}`)
			let mediaFile = findOTRMedia(otr.media)
			if(mediaFile){
				otrMedia = af.addMediaFile(mediaFile);	
				step.file(mediaFile,`Found media file`);
			}
		}
		
		const tiers = {};
		const annotations = [];
		// for oTranscribe Jsoup seems best
		// but tbh otr contentEditable is not machine friendly,
		// so we must parse it with hacks?
		let html = otr.text;
		html = html.split('<p>').join('');
		html = html.split('</p>').join('\n');
		html = html.split('<br />').join('\n');
		const tspan = '<span class="timestamp"';
		let p = html.indexOf(tspan)
		while(p>-1){
			let timestamp = html.substring(p,html.indexOf('</span>',p+1)).split('>')[1]			
			let exactTime = html.substring(p,html.indexOf('>',p+1)).split('data-timestamp=')[1];
			exactTime = exactTime.split('"').join('');
			
			// then the value is until next timestamp:
			let p2 = html.indexOf(tspan,p+1);			
			let value = html.substring(html.indexOf('</span>',p+1)+7, p2<0 ? html.length-1 : p2);
			
			// might have leftovers contenteditable
			value = value.split('<span>').join('');
			value = value.split('</span>').join('');
			value = value.trim();
			
			//step.info(timestamp+':'+exactTime+"="+value);
			let tier = 'otranscribe';
			if(tierSeparator){
				let a = value.split(tierSeparator);
				if(a.length>1){
					tier = a.shift().trim();					
					value = a.join(tierSeparator).trim();
				}
			}
			if(!tiers[tier]) tiers[tier] = af.addTier({id:tier});
						
			// ok we'll assume an annotation lasts as long as until the next one
			annotations.push({
				tier:tiers[tier],
				id:exactTime,
				start:exactTime*1000,
				value:value
			});						
			p = p2;
		}		
		// so it means we have to calculate their duration now		
		for(let i = 0;i<annotations.length;i++){
			let current = annotations[i];
			let next = annotations[i+1];			
			current.stop = next ? next.start - 1 : (otrMedia ? otrMedia.getDuration() : next.start + 1000)
			af.addAnnotation(current);
		}
		step.ok();
		step.info(`Found ${Object.keys(tiers).length} tiers and ${Object.keys(annotations).length} annotations`);
		return af;
	}
}