/** 
@author CG
@log 2024-02-22 migrated to provider system
@label add a SRT file
@hidden
@attr file file () The .SRT file to load
@attr media file () The associated media file
@attr encoding string (autodetect) The encoding to use when loading the SRT file
@attr tier string (srt) A tier name to link to the SRT annotations
@attr group string () The group to associate to this corpus file and all its annotations
@attr tags tags () A list of tags associated to this corpus file
Provides domain mapping for the SubRip annotation format
https://en.wikipedia.org/wiki/SubRip

@dependency human-file-size.js
@register extension srt
**/

provider.subrip = {
	load:function(){
		const step = log.addStep({name:`Loading SRT`});
		if(!attr.file){
			step.error(`No .SRT file was specified`);
			return;
		}
				
		let file = new java.io.File(attr.file);
		if(!file.exists()){
			step.error(`File "${attr.file}" does not exist`);
			return;
		}
		
		let charset;
		if(attr.encoding && attr.encoding!='autodetect'){
			try {
				charset = java.nio.charset.Charset.forName(attr.encoding);
				step.info("User charset: "+charset.name());
			}catch(ex){
				step.error("Unknown charset: "+attr.encoding);
			}
            
        }
		if(!charset){
            charset = org.avaatoolkit.util.CharsetDetector.detect(file);
            step.info("Detected charset: "+charset.name());
        }		
		step.info(`Loading file with charset "${charset.name()}"`);
		let content = api.readFile(file,charset);
		if(!content){
			step.error(`Could not read file ${attr.file}`);
			return;
		}        
        step.ok();
		// create a generic AnnotationFile instance
		let af = new org.avaatoolkit.formats.GenericFile(api.doc, file);		
		af.group = attr.group || '';	
		if(attr.tags) af.tags.addAll(attr.tags.split(','));		
		af.domainKey = domainKey; // must pass current special domainKey variable //TODO refactor possible?
		api.doc.domain.add(af);
		
		// the tier that will hold the annotations
		let tier = af.addTier({id:attr.tier || 'srt'});
        
		// parsing srt
		let lines = content.split("\n");
        for (let i = 0; i < lines.length; i++) lines[i] = lines[i].split("\r").join("");
        for (let i = 0; i < lines.length; i+=0) {
            let id = lines[i];
			let codeLine = lines[i + 1];
			if(!id && !codeLine) break;
			if(!codeLine){
				step.error(`Missing timecode line after id "${id}" line ${i}`);
				break;
			}
            let times = codeLine.substring(0,29);
            let codes = times.split("-->");
            if (codes.length != 2) {
				step.error(`Can't parse srt time codes: ${times}`);
				break;
			}
            let start = parseTimeCode(codes[0]);
            let stop = parseTimeCode(codes[1]);
            let value = "";
            let j = 0;
            while (true) {
                j++;
                let p = i + 1 + j;
                if (p >= lines.length) break;
                let lineOfText = lines[p];
                if (!lineOfText) break;
                value += !value ? lineOfText : "\n" + lineOfText;
            }
            i += 2 + j;
			// insert annotation into domain file
			af.addAnnotation({
				tier:tier,
				id:id,
				start:start,
				stop:stop,
				value:value
			});
        }
		if(!af.annotations.size()){
			step.warn(`No annotation found in the file!`);
		}else{
			step.info(`Found ${af.annotations.size()} annotations`);
			step.ok();
		}	
		
		// now let's check for the media file
		let mediaFile;
        if(attr.media){
			mediaFile = new java.io.File(attr.media);
			if(!mediaFile.exists()){
				step.warn(`File "${attr.media}" does not exist`);				
				mediaFile = null;
			}else{				
				step.info(`Media file exists`);
				step.ok();
			}
        }
		const exts = ["mp4","avi","mkv","mov","divx","xvid"];
		if(!mediaFile){ // trying autodetect						
			let baseName = org.apache.commons.io.FilenameUtils.getBaseName(file.getName());			
			let parent = file.getParent();			
			let allExts = exts.concat(exts.map(e=>e.toUpperCase()));
			for(let ext of exts){
                let f = new java.io.File(parent+"/"+baseName+"."+ext);
                if(f.exists()){                    
                    mediaFile = f;
					step.info(`Found media file with same name`);	
					step.ok();
                    break;
                }
            }
		}	
		if(!mediaFile){ // ok, one more try			
			let mediaFiles = [];
			let allFiles = file.getParentFile().listFiles();						
			for(let f of allFiles){
				let ext = org.apache.commons.io.FilenameUtils.getExtension(f.getName());				
				if(exts.indexOf(ext)>-1) mediaFiles.push(f);									
			}
			if(mediaFiles.length>1){
				step.warn(`Too many media files to chose from:`)
				mediaFiles.forEach(f=>step.file(f));
			}
			if(mediaFiles.length===1){
				step.info(`Found a media file in same folder`)
				step.ok();
				mediaFile = mediaFiles[0];
			}
		}
		if(mediaFile){			
			af.addMediaFile(mediaFile);
		}else{
			step.error(`Could not find associated media file`);
		}
		return af;
	}
		
		
		
		
	
}


function parseTimeCode(code){
	code = code.trim();
	t = code.split(",");
	ms = t[1]*1;
	t = t[0].split(":");
	h = t[0]*1;
	m = t[1]*1;
	s = t[2]*1;
	return (h*3600+m*60+s) * 1000 + ms;
}