/** 
@author CG
@input Array<Annotation>
@attr show-value bool (true) Whether to display the annotation's text value
@attr show-video bool (false) Whether to display the video clip of the processed annotation
@attr show-tier bool (false) Whether to display the annotation's tier
@attr show-group bool (false) Whether to display the annotation's group
@attr show-file bool (false) Whether to display the annotation's file
@attr show-index bool (false) Whether to display the index of each row
@attr video-with-cotext bool (true) Whether to include cotext annotations in the video
@attr cotext int (1) Number of annotations to extract before and after the processed annotation
@attr cotext-before int (cotext) Number of annotations to extract before the processed annotation
@attr cotext-after int (cotext) Number of annotations to extract after the processed annotation
@attr cotext-video bool (false) Whether to display a video for cotext annotations
@attr cotext-split bool (false) Whether to display each cotext annotation in its own column
@todo cotext-time bool (false) Whether to display the time difference between cotext annotations
@attr max-time-diff int (60) Skip cotext annotations that are out of this range

@attr label-before string (Before) Label for table header
@attr label-pivot string (Pivot) Label for table header
@attr label-after string (After) Label for table header
@attr label-index string ( ) Label for table header
@attr label-tier string (Tier) Label for table header
@attr label-group string (Group) Label for table header
@attr label-file string (File) Label for table header
@attr label-none string ( ) Label for no matches

@log 2023-01-03 initial implementation
@log 2023-02-16 added cotext-video, cotext-split, video-with-cotext
@log 2024-06-19 added show-tier show-group show-file show-index

The concordancer view displays a table of annotations and their cotext annotations.
Attributes allow timerange and count limits, to extract only meaningful relative annotations.
Configurable display to see each cotext annotation in its own column, or all combined into one clip.

@sample VIEW max-time-diff="15" cotext="1"
@image jpg

@sample VIEW max-time-diff="15" cotext="2" show-video="true" video-with-cotext="false" cotext-split="true" cotext-video="true"
@image split.jpg

**/
view.concordancer = {
	list:list=>{				
		var cotext = attr.cotext*1 || 1;
		var cotextBefore = attr['cotext-before']*1 || cotext;
		var cotextAfter = attr['cotext-after']*1 || cotext;
		var cotextVideo = attr['cotext-video'] ? attr['cotext-video']=='true' : false;
		var cotextSplit = attr['cotext-split'] ? attr['cotext-split']=='true' : false;
		var cotextTime = attr['cotext-time'] ? attr['cotext-time']=='true' : false;
		var maxTimeDiff = attr['max-time-diff']*1000 || 60*1000;
		var showVideo = attr['show-video'] ? attr['show-video']=='true' : false;
		const showTier = attr['show-tier'] ? attr['show-tier']=='true' : false;
		const showGroup = attr['show-group'] ? attr['show-group']=='true' : false;
		const showFile = attr['show-file'] ? attr['show-file']=='true' : false;
		const showIndex = attr['show-index'] ? attr['show-index']=='true' : false;
		var showVideoWithCotext = attr['video-with-cotext'] ? attr['video-with-cotext']=='true' : true;
		var showValue = attr['show-value'] ? attr['show-value']=='true' : true;		
		var asort = (a,b)=>a.start>b.start?-1:1;
		var s = `<table class="${cotextSplit?'split':''}"><thead><tr>`;		
		if(showIndex) s+=`<td>${esc(attr['label-index']||'')}</td>`;
		if(showGroup) s+=`<td>${esc(attr['label-group']||'Group')}</td>`;
		if(showFile) s+=`<td>${esc(attr['label-file']||'File')}</td>`;
		if(showTier) s+=`<td>${esc(attr['label-tier']||'Tier')}</td>`;
		
		var colspanBefore = cotextSplit ? ` colspan="${cotextBefore}"` : '';
		var colspanAfter = cotextSplit ? ` colspan="${cotextAfter}"` : '';
		var vid = a=>s+=`<video preload="none" poster="${a.snap()}" controls width="300"><source src="${a.clip()}" type="video/mp4"></video>`;		
		s+=`<td${colspanBefore} class="before">${esc(attr['label-before']||'Before')}</td>`;
		s+=`<td class="current pivot">${esc(attr['label-pivot']||attr['label-current']||'Pivot')}</td>`;
		s+=`<td${colspanAfter} class="after">${esc(attr['label-after']||'After')}</td>`;
		s+=`</tr></thead><tbody>`;
		var index = 0;
		list.forEach(a=>{
			var an = a.tier.annotations.toArray();
			var before=[];
			an.forEach(aa=>{
				if(aa.start < a.start){
					if(before.length<cotextBefore) before.push(aa);
					else {
						before.sort(asort);
						if(aa.start > before[cotextBefore-1].start) before[cotextBefore-1]=aa;
					}
					before.sort(asort);
				}
			});
			var after=[];
			an.forEach(aa=>{
				if(aa.start > a.start){					
					if(after.length<cotextAfter) after.push(aa);
					else {
						after.sort(asort).reverse();
						if(aa.start < after[cotextAfter-1].start) after[cotextAfter-1]=aa;
					}
					after.sort(asort).reverse();
				}
			});
			before.reverse(); //wtf
			if(maxTimeDiff){				
				while(before.length){
					if(before[0].start < a.start - maxTimeDiff) before.shift();
					else break;
				}
				while(after.length){
					if(after[after.length-1].start > a.stop + maxTimeDiff) after.pop();
					else break;
				}
			}
			s+=`<tr>`;
			if(showIndex) s+=`<td>${++index}</td>`;
			if(showGroup) s+=`<td>${a.af.group}</td>`;
			if(showFile) s+=`<td>${a.af.id}</td>`;
			if(showTier) s+=`<td>${a.tier.id}</td>`;
			if(cotextSplit){ // one column per cotext annotation
				var A = before.slice().reverse();
				for(var i=0;i<cotextBefore;i++){
					s+=`<td class="before">`;
					var aa = A[cotextBefore-(i+1)];
					if(aa){
						if(cotextVideo) vid(aa);
						if(showValue) s+=`<label class="annotation">${esc(aa.value || '')}</label>`;
					}
					s+=`</td>`;
				}
			}else{ // all annotations in one column
				s+=`<td class="before">`;
				if(cotextVideo && before.length){
					var atmp = a.clone(); // build temporary annotation to combine cotext annotations
					atmp.start = before[0].start;
					atmp.stop = before[before.length-1].stop;
					vid(atmp);
				}
				if(showValue && before.length) before.forEach(aa=>{					
					s+=`<label class="annotation">${esc(aa.value)}</label>`;
				}); else s+=`<i>${attr['label-none']||''}</i>`;				
				s+=`</td>`;								
			}
			s+=`<td class="current pivot">`;
			if(showVideo){
				var atmp = a.clone();
				if(showVideoWithCotext){
					var atmp = a.clone(); // build temporary annotation
					if(before.length) atmp.start = before[0].start;
					if(after.length) atmp.stop = after[after.length-1].stop;
					vid(atmp);
				}else{					
					vid(a);
				}
			}
			if(showValue) s+=`<label class="annotation">${esc(a.value)}</label>`;
			s+=`</td>`;
			if(cotextSplit){				
				for(var i=0;i<cotextAfter;i++){
					s+=`<td class="after">`;
					var aa = after[i];
					if(aa){						
						if(cotextVideo) vid(aa);
						if(showValue) s+=`<label class="annotation">${esc(aa.value || '')}</label>`;
					}
					s+=`</td>`;
				}
			}else{			
				s+=`<td class="after">`;
				if(cotextVideo && after.length){
					var atmp = a.clone();
					atmp.start = after[0].start;
					atmp.stop = after[after.length-1].stop;
					vid(atmp);
				}
				if(showValue && after.length) after.forEach(aa=>{					
					s+=`<label class="annotation">${esc(aa.value)}</label>`;
				}); else s+=`<i>${attr['label-none']||''}</i>`;				
				s+=`</td>`;
			}
			s+=`</tr>`;
		});
		s+=`</tbody></table>`;
		return s;
	},
	css:s=>{
		return `
#view video {
	border-radius:4px;	
}
#view table {
	border-collapse: collapse;
}
#view table.split thead {
	text-align:center;
}
#view table.split td.pivot {
	border-left: 2px solid gray;
	border-right: 2px solid gray;
}
#view td {
	vertical-align: top;
	border: 1px solid gainsboro;
	padding: 5px;
}
#view video {
	display:block;
}
#view label {
    display: block;
}
`;
	}
}