/** 
@author CG
@input Array<Annotation>
@attr group bool (false) Whether to separate annotations based on their group instead of their file
@attr sequence string Separate annotations based on another tier's annotations (used as sequences)
@attr time-offset bool (false) Whether to use files' time-offset
@attr time-collapse number (0) Number of seconds after which empty space is collapsed (0 = no collpase)
@attr zoom-factor int (100) Milliseconds to pixel factor
@deprec @attr colors string A comma separated list of colors, like "#c7f464,#ffa822"
@attr show-alt-tiers bool (false) Whether to display alternative tiers in the main table
@attr show-alt-tier-in-empty-cells bool (false) Whether to display the alternative tier if the cell is empty
@attr show-value-on-hover bool (false) Whether to display the annotation value on mouse over
@attr show-snap-on-hover bool (false) Whether to display the video snapshot on mouse over
@attr show-video-on-click bool (false) Whether to display the video player on click
@attr video-player-width string (300)
@attr video-player-height string ()
@param tier="id" name="" parent="" color="" alt-tier="" | Customize tier display
@param group="id" name="Custom Name" | Set a custom header name for each group
@param file="id" name="Custom Name" | Set a custom header name for each file
@param alt-tier="id" label="" css="" | Customize alternative tier display and legend
@param !name
@param !label
@param !parent
@param !color
@param !css
@param !alt-tier
@log 2023-03-05 initial implementation
@log 2023-06-13 added customizable tier parent and color

The density view plots annotations as filled bars in a timeline, revealing interactions frequency and duration between tiers.

@sample VIEW zoom-factor="500"
@image png
**/
view.density = {
	list:list=>{		
		var colors = ['#556270', '#4ecdc4', '#c7f464', '#ff6b6b', '#c44d58','#ffa822', '#134e6f', '#ff6150', '#1ac0c6', '#dee0e6'];
		var factor = attr['zoom-factor']*1 || attr.factor*1 || 100;
		if(attr.colors){
			var c = "['"+attr.colors.split(',').join("','")+"']";
			try {
				colors = JSON.parse(c);
			}catch(ex){
				log.error('failed to parse `colors` attribute');
				log.error(c);
				log.error(ex.message);
			}
		}
		var addpx = (v,d) => v && (v*1===v*1) ? v+'px' : (d ? d+'px' : '');
		var defaultNoDataCSS = 'background:repeating-linear-gradient( 45deg, #000000, #000000 3px, #ffffff00 3px, #ffffff00 6px)';
		var color = 'black';		
		var group = !attr.group ? false : (attr.group == 'true'); 
		var showValueOnHover = !attr['show-value-on-hover'] ? false : (attr['show-value-on-hover'] == 'true'); 
		var showSnapOnHover = !attr['show-snap-on-hover'] ? false : (attr['show-snap-on-hover'] == 'true'); 
		var showVideoOnClick = !attr['show-video-on-click'] ? false : (attr['show-video-on-click'] == 'true'); 
		var showNoDataTiers = !attr['show-alt-tiers'] ? false : (attr['show-alt-tiers'] == 'true'); 
		var showNoDataInEmptyCell = !attr['show-alt-tier-in-empty-cells'] ? false : (attr['show-alt-tier-in-empty-cells'] == 'true'); 
		var videoPlayerWidth = addpx(attr['video-player-width'],300);
		var videoPlayerHeight = addpx(attr['video-player-height']);
		var sequence = attr.sequence ? attr.sequence : '';		
		var tierNames = {}; // map of custom names from options
		var groupNames = {};
		var fileNames = {};
		var asort = (A,B)=>{ return A.start > B.start ? 1 : -1};
		var tiers = {}; // map of tier id => tier's annotations			
		var files = {}; // map of file name => file's annotations
		var groups = {}; // ... 
		var sequences = {}; // ...
		var tierParents = {};
		var tierColors = {};
		var tierNoData = {};
		var noDataTiers = {};
				
		params.forEach(opt=>{ 
			if(opt.hasAttribute('tier')){
				var tierid = opt.getAttribute('tier');
				tierNames[tierid] = opt.getAttribute('name');
				tiers[tierid] = []; // force order
				if(opt.hasAttribute('parent') && opt.getAttribute('parent')){
					tierParents[tierid] = opt.getAttribute('parent');
				}
				if(opt.hasAttribute('alt-tier') && opt.getAttribute('alt-tier')){
					tierNoData[tierid] = opt.getAttribute('alt-tier');
				}
				if(opt.hasAttribute('color') && opt.getAttribute('color')){
					tierColors[tierid] = opt.getAttribute('color');
				}
				if(!tierColors[tierid]){ // let's try loading from global colors
					let c = api.doc.style.colorOfTier(tierid);
					if(c) tierColors[tierid] = c;
				}
				return;
			}
			if(opt.hasAttribute('alt-tier') && opt.getAttribute('alt-tier')){
				var tierid = opt.getAttribute('alt-tier');
				var o = noDataTiers[tierid] = {id:tierid};
				if(opt.hasAttribute('color') && opt.getAttribute('color')) o.color = opt.getAttribute('color');
				if(opt.hasAttribute('css') && opt.getAttribute('css')) o.css = opt.getAttribute('css');
				if(opt.hasAttribute('name') && opt.getAttribute('name')) o.name = opt.getAttribute('name');
				if(opt.hasAttribute('label') && opt.getAttribute('label')) o.label = opt.getAttribute('label');
				return;
			}
			if(opt.hasAttribute('group')){
				groupNames[opt.getAttribute('group')+''] = opt.getAttribute("name")+'';
			}
			if(opt.hasAttribute('file')){
				fileNames[opt.getAttribute('file')+''] = opt.getAttribute("name")+'';
			}
		});
		
		if(sequence) group = false; // AF based		
		list.forEach(a=>{
			if(!tiers[a.tier.id]) tiers[a.tier.id] = [];
			tiers[a.tier.id].push(a)
			if(!tiers[a.tier.id].color) tiers[a.tier.id].color = tierColors[a.tier.id] ? tierColors[a.tier.id] : (colors.length ? colors.shift() : 'black');
			if(!files[a.af.id]){
				files[a.af.id] = [];
				files[a.af.id].id=a.af.id;
			}
			var	f = files[a.af.id];			
			f.push(a);
			if(f.start===undefined || a.start<f.start){				
				//log.info('SETTING start of '+f.id+' to '+a.start);
				f.start = a.start;
			}
			if(!groups[a.af.group]) groups[a.af.group] = [];
			groups[a.af.group].push(a);

			if(sequence){
				a.af.getTier(sequence).annotations.forEach(seq=>{
					if(a.start>=seq.start && a.stop<=seq.stop){
						var sid = a.af.id+'/'+seq.value;
						if(!sequences[sid]){
							var s = sequences[sid] = [];
							s.id=sid;
							s.name = seq.value+'';
							s.af = seq.af;
							s.start = seq.start;
							s.stop = seq.stop;							
						}
						sequences[sid].push(a);
					}
				});
			}
		});
		var prev;		
		var aset = group ? groups : files;
		//log.info(Object.keys(aset));
		if(sequence) {
			aset = {};
			var a = [];
			for(var k in sequences) a.push(sequences[k]);
			a.sort(
				(a,b)=>{ return a.af.filename+"" == b.af.filename+"" ?  (a.start > b.start ? 1 : -1) : (a.af.filename+"" < b.af.filename+"" ? -1 : 1)}
			);
			a.forEach(seq=>aset[seq.id]=seq);
		}
		var names = group ? groupNames : fileNames;
		
		var hasParents = Object.keys(tierParents).length > 0;
		var s = `<table><thead><tr>`;
		if(hasParents) s+= `<td class="corner" colspan="2"></td>`;
		else s+= `<td class="corner"></td>`;
		for(key in aset){
			var kname = sequence ? aset[key].name : names[key] || key;
			s += `<td class="af-separator sequence-separator">${esc(kname)}</td>`;
		}
		s += `</tr></thead>`;
		
		s += `<tbody>`;	
		var didParent = {};
		var countParent = {};
		if(hasParents) for(tier in tiers){
			var par =  tierParents[tier] || '';
			if(!countParent[par]) countParent[par]=1;
			else countParent[par]++;
		}
		
		function draw(tier,alist,classes,toggleOpacity,combine){
			classes = classes || '';
			var last = 0;
			var opacity = 1;
			alist.sort(asort);
			alist.forEach(a=>{
				if(a.tier.id+'' != tier) return;
				//if(prev == a) return; // duplicates from possible OPerations
				//prev = a;										
				var start = a.start - alist.start;
				var stop = a.stop - alist.start;
				var dur = stop-start;
				var F = Math.ceil;
				var w = F(dur/factor);
				var left = F(start/factor);
				if(last>0){
					var b = F((start-last)/factor);
					s+=`<div class="blank" style="width:${b}px"></div>`;
				}else if(start>0){
					var b = F((start)/factor);
					s+=`<div class="blank" style="width:${b}px"></div>`;
				}
				
				s+=`<div class="annotation-pad" title="${esc(a.value)}" style="width:${w}px;background-color:${tiers[a.tier.id+''].color||'red'};"></div>`;
				s+=`<div class="annotation ${classes}" title="${esc(a.value)}" style="opacity:${opacity};left:${left}px;width:${w}px;background-color:${tiers[a.tier.id+''].color||'red'};"></div>`; 
				last = stop;
				if(toggleOpacity) opacity = opacity == 1 ? 0.8 : 1;
				
			});
		}
		function drawNoData(tier,alist){
			if(!alist.length) return;			
			var last = 0;
			var opacity = 1;			
			alist.sort(asort);		
			var toDraw = [];			
			alist.forEach(a=>{				
				if(a.tier.id+'' != tier) return;
				if(!toDraw.length){
					toDraw.push(a.clone());
					return;
				}
				var lastA = toDraw[toDraw.length-1];							
				
				if(lastA.stop==a.start){
					//log.info('COMBINE at '+a.start);
					lastA.stop = a.stop;
					return;
				}
				else{
					var diff = a.start-lastA.stop;
					if(diff<factor){
						//log.info('COMBINE SMALL DIFF '+diff);
						lastA.stop = a.stop;
						return;
					}
					
				}
				toDraw.push(a.clone());
			});
			toDraw.forEach(a=>{				
				var start = a.start - alist.start;
				var stop = a.stop - alist.start;				
				var dur = stop-start;
				var F = Math.ceil;
				var w = F(dur/factor);
				var left = F(start/factor);
				if(last>0){
					var b = F((start-last)/factor);
					s+=`<div class="blank" style="width:${b}px"></div>`;
				}else if(start>0){
					var b = F((start)/factor);
					s+=`<div class="blank" style="width:${b}px"></div>`;
				}
								
				var css = noDataTiers[tier] && noDataTiers[tier].css ? noDataTiers[tier].css : defaultNoDataCSS;
				
				s+=`<div class="annotation-pad" title="${esc(a.value)}" style="width:${w}px;background-color:${tiers[a.tier.id+''].color||'red'};"></div>`;
				s+=`<div class="annotation stripes" title="${esc(a.value)}" style="opacity:${opacity};left:${left}px;width:${w}px;${css}"></div>`;
				last = stop;				
			});
		}
		
		var isNoData = {}
		for(var k in tierNoData){
			isNoData[tierNoData[k]] = true;
		}
		
		for(tier in tiers){
			
			if(isNoData[tier]){
				if(!showNoDataTiers) continue;
			}
			
			var tname = tierNames[tier] || tier;
			s += `<tr class="tier">`;
			if(hasParents){
				var par =  tierParents[tier] || '';
				if(!didParent[par]){
					didParent[par]=true;
					s += `<td rowspan="${countParent[par]}" class="parent-name">${par}</td>`;
				}
					
			}
			s += `<td class="tier-name">${esc(tname)}</td>`;
			var absoluteDivs = [];
			for(key in aset){				
				//var kname = sequence ? aset[key].name : names[key] || key;						
				s += `<td class="tier-density ${showVideoOnClick ? 'clickable' : ''}">`;
				var finalw = 0;
				var last = 0;
				var opacity = 1;
				aset[key].sort(asort);				
				aset[key].forEach(a=>{
					if(a.tier.id+'' != tier) return;
					if(prev == a) return; // duplicates from possible OPerations
					//log.info(a.id);
					prev = a;										
					var start = a.start - (aset[key].start || 0);					
					var stop = a.stop - (aset[key].start || 0);
					var dur = stop-start;
					var F = Math.ceil;
					var w = F(dur/factor);
					var left = F(start/factor);
					if(isNaN(left)){
						log.error(aset[key].start);
						throw "game over.";
					}
						
					if(last>0){
						var b = F((start-last)/factor);
						s+=`<div class="blank" style="width:${b}px"></div>`;
					}else if(start>0){
						var b = F((start)/factor);
						s+=`<div class="blank" style="width:${b}px"></div>`;
					}
					/*var b64v="";
					try {
						b64v = atob(esc(String(a.value)));
					}catch(ex){
						log.warn("failed b64 on "+a.value)
					}*/
					s+=`<div class="annotation-pad" title="${esc(a.value)}" style="width:${w}px;background-color:${tiers[a.tier.id+''].color||'red'};"></div>`;
					s+=`<div class="annotation" title="${esc((a.value))}"`;
					if(showSnapOnHover){
						s+=` data-snap-src="${esc(a.snap())}"`;
					}
					if(showValueOnHover){
						s+=` data-value="${esc(a.value)}"`; //TODO b64
					}
					if(showVideoOnClick){
						s+=` data-clip-src="${esc(a.clip())}"`;
					}					
					s+=` data-player-width="${videoPlayerWidth}" data-player-height="${videoPlayerHeight}"`;
					s+=` style="opacity:${opacity};left:${left}px;width:${w}px;background-color:${tiers[a.tier.id+''].color||'red'};"></div>`;
					last = stop;
					opacity = opacity == 1 ? 0.8 : 1;
					finalw = left + w;
				});
				
				if(tierNoData[tier]) {
					if((showNoDataInEmptyCell && !finalw) || finalw){
					//s+=`<i>${tierNoData[tier]}</i>`;
						s+=`<div class="nodata-nooverflow" sstyle="width:${finalw}px"><div>`;
						drawNoData(tierNoData[tier],aset[key]);
						s+=`</div></div>`;
					}
				}
				
				s += `</td>`;
			}
			//s += absoluteDivs.reverse().join('');	
		
		}
		
		s += `</tbody></table>`;
		if(Object.keys(noDataTiers).length){
			s += `<div class="density-legend">`;
			for(var k in noDataTiers){
				var o = noDataTiers[k];
				var css = o.css ? o.css : defaultNoDataCSS;
				s += `<div class="alt-tier-legend"><span class="alt-tier-block" style="${css}"></span><label>${o.label||o.name||k}</label></div>`;
			}
			s += `</div>`;
		}
		return s;		
	},
	js:s=>{
		$(ready=>{
			$('.tier-density .annotation').each((i,e)=>{
				var $e = $(e);
				var snap = $e.data('snap-src');		
				var value = $e.data('value');
				//if(value) value = btoa(value);
				if(snap) Tipped.create(e,'<div><img style="max-width:700px" src="'+snap+'"/></div>');
				if(value) Tipped.create(e,'<div>'+value+'</div>');
				var clip = $e.data('clip-src');
				var pw = $e.data('player-width');
				var ph = $e.data('player-height');
				if(clip) Tipped.create(e,`<div class="density-popup-video" style="${pw?'width:'+pw+';':''}${ph?'height:'+ph+';':''}"><video preload="none" autoplay="true" poster="" controls onplay="Tipped.refresh(this)"><source src="${clip}" type="video/mp4"></video></div>`,
					{showOn:'click', hideOn:false, hideOnClickOutside: true,			
					 onShow: function (content, element) {
					  $(element).addClass("highlight");
					},
					afterHide: function (content, element) {
					  $(element).removeClass("highlight");
					}
					});
			});
						
		});
	},
	css:s=>{
		return `
.density-popup-video {
	pointer-events:all;
}
.density-popup-video video {
	width:100%;
	height:100%;
}
#view td {
    border: 1px solid gainsboro;
}
#view td.tier-density.clickable .annotation {
	cursor:pointer;
}
#view td.tier-density {
	position:relative;
}
#view td .annotation {
	position: absolute;
    top: 0px;
    height: 48px;    
}

#view td .annotation.highlight {
	border: 2px dotted;
}
#view .annotation-pad {
    height: 40px;    
    display: inline-block;	
	visibility:hidden;
}
#view tr.tier td.tier-name,
#view tr.tier td.parent-name {
    white-space: nowrap;
    padding: 10px;
}
#view .sequence-separator {
    text-align: center;
    font-weight: bold;
    padding-left: 10px;
    padding-right: 10px;
}
#view td.tier-density {
    padding-right: 20px;
    padding-top: 3px;
    padding-bottom: 0px;
    padding-left: 0px; 
}
#view td.corner {
	border-top:none;
	border-left:none;
}
#view table {
    border-collapse: collapse;
}
#view .tier-density {
    white-space: nowrap;
}
#view .blank {
    height: 40px;    
    display: inline-block;
}

#view tr.time-separator {	
    height: 50px;
	background: rgb(255,255,255);
	background: linear-gradient(0deg, rgba(255,255,255,0) 0%, rgba(0,0,0,0.08167016806722693) 50%, rgba(255,255,255,0) 100%);
	text-align:center;
}
#view tr.tier-header {	
    font-weight:bold;
	text-align:center;
}
#view tr.annotation {
    vertical-align: top;
	
}
#view tr.annotation td:first-child {
    color:gray;
	font-size: 14px;
    padding-right: 5px;
}
#view tr.af-separator td {
    background: black;
    color: white;
    font-weight: bold;
    padding: 5px;
    border-radius: 8px;
    text-align: center;
}
#view tr.annotation td {
    border-top: 1px dashed gainsboro;
    max-width: 320px;
	padding-top: 10px;

}
#view video {
    border-radius: 6px;
	display: block;
}
#view .nodata-nooverflow > div {
	position:relative;
}
#view .nodata-nooverflow {
    position: absolute;
    top: 0px;
    left: 0px;
    overflow: hidden;
	width: -webkit-fill-available;
	width: -moz-available;
	height: -webkit-fill-available;
	height: -moz-available;
	pointer-events: none;
}
#view .alt-tier-block {
    height: -webkit-fill-available;
	height: -moz-available;
}
#view .alt-tier-legend {
	margin:5px;
}
#view .alt-tier-legend span {
    display: inline-block;
    width: 50px;
    height: 18px !important;
    margin-right: 5px;
	border-radius:6px;
}
#view .alt-tier-legend label {
	vertical-align: top;
	font-style:italic;
}

#view .density-legend {
    padding-top: 10px;
}
`

	}
}

function fdur(s){
    var sec_num = parseInt(s, 10);
    var hours   = Math.floor(sec_num / 3600);
    var minutes = Math.floor((sec_num - (hours * 3600)) / 60);
    var seconds = sec_num - (hours * 3600) - (minutes * 60);    
	if(!hours) {
		if(!minutes) return seconds+" seconds";
		return minutes+" minutes";		
	}
    return hours+' hours';
}

function ftime(s){
    var sec_num = parseInt(s, 10);
    var hours   = Math.floor(sec_num / 3600);
    var minutes = Math.floor((sec_num - (hours * 3600)) / 60);
    var seconds = sec_num - (hours * 3600) - (minutes * 60);
    //if (hours   < 10) {hours   = "0"+hours;}
    if (minutes < 10) {minutes = "0"+minutes;}
    if (seconds < 10) {seconds = "0"+seconds;}
    return hours+':'+minutes+':'+seconds;
}