/** 
@author CG
@input Array<Annotation>
@attr reference tier The tier used as reference, if any. 
@deprec @attr columns string Comma separated list of tier names, to set column order
@deprec @attr show-value bool (true) Whether to display the annotation's text value (makes no sense with load-annotations-from-forms::extract-original-annotations)
@attr show-video bool (true) Whether to display the annotation's video clip
@attr show-index bool (true) Whether to display the index
@attr values-replacement js () A JSON object used to replace a given value with another one, like {"a1":"A", "a2":"A"}
@attr weight-matrix bool (false) Whether to use the weight matrix
@attr weight-matrix-values matrix () A custom matrix defining the weight of each values' pair
@attr weight-matrix-func js () A custom JS function that should return the weight for a given pair provided as arguments, like (a,b) => a==b ? 1 : 0
@attr save-weight-matrix-csv string () Name of a CSV file to export the weight matrix
@attr save-coincidence-matrix-csv string () Name of a CSV file to export the coincidence matrix
@attr save-krippendorff-alpha-json string () Name of a JSON file to export Krippendorff's alpha, as {"alpha":X}
@todo group bool (false)  Whether to separate annotations based on their group instead of their EAF file
@todo sequence string Separate annotations based on another tier's annotations (used as sequences)
@todo sequence-relative-time bool (false) Whether to display time relative to sequence start
@param column="tier id" name="" | Set custom column order and header names
@param exclude-tag="tag" | Exclude from the intercoding calculations all annotations that have this tag 
@param name-of-tier="id" name="Custom Name" | Set a custom header name for each tier
@param name-of-group="id" name="Custom Name" | Set a custom header name for each group
@param name-of-file="id" name="Custom Name" | Set a custom header name for each file
@param +stat="score" name="Score (Points)" | Add a stat column (score alone)
@param +stat="score%" name="Score (Percentage)" | Add a stat column (score percentage)
@param +stat="score/" name="Score (Points/Total)" | Add a stat column (score on total)
@param +stat="count" name="Similarity percentage count"  | Add a stat column
@param +stat="weight-matrix" name="Weight Matrix" | Show the weight matrix
@param +stat="coincidence-matrix" name="Coincidence Matrix" | Show the coincidence matrix
@param +stat="krippendorff-alpha" name="Krippendorff's Alpha" | Show Krippendorff's Alpha
@param !name
@param !stat
@image png
@log 2023-02-21 initial implementation

The intercoding view makes it easy to process JSON files resulting from forms, and display them in a meaningful way for intercoding validation and statistics.

**/
view.intercoding = {
	list:list=>{		
		var collapse = attr.collapse ? attr.collapse*1 || 60 : 60; 
		var showVideo = attr['show-video'] ? attr['show-video'] == 'true' : true;
		var showValue = attr['show-value'] ? attr['show-value'] == 'true' : true; 		
		var group = !attr.group ? false : (attr.group == 'true'); 
		var showStart = (attr['show-start']=='true'); 
		var showIndex = !(attr['show-index']=='false'); 
		var sequence = attr.sequence ? attr.sequence : '';		
		var tierNames = {}; // map of custom names from options
		var useWeightMatrix = attr['weight-matrix']=='true';
		var saveWeightMatrixCSV = attr['save-weight-matrix-csv'];
		var saveCoincidenceMatrixCSV = attr['save-coincidence-matrix-csv'];
		var saveKrippendorffAlphaJSON = attr['save-krippendorff-alpha-json'];
		var weightMatrixFunc = attr['weight-matrix-func'];
		var valuesReplacement = null;
		if(attr['values-replacement']){
			try {
				valuesReplacement = JSON.parse(attr['values-replacement']);			
			}catch(e){
				log.error(`Could not parse values-replacement JSON`);
			}
		}
		const valueReplace = v=>valuesReplacement && valuesReplacement[v]!==undefined?valuesReplacement[v]:v;
		const roundec = v=>parseFloat(v.toFixed(4));
		var groupNames = {};
		var fileNames = {};
		var statCols = [];
		var excludeTags = ['exclude-from-intercoding']; // all annotations with these tags won't be used for intercoding, only display
		var columns = [];
		var globalStats = {'krippendorff-alpha':false,'coincidence-matrix':false,'weight-matrix':false};
		params.forEach(opt=>{ 
			if(opt.hasAttribute('name-of-tier')){
				tierNames[opt.getAttribute('name-of-tier')+''] = opt.getAttribute("name")+'';
			}
			if(opt.hasAttribute('name-of-group')){
				groupNames[opt.getAttribute('name-of-group')+''] = opt.getAttribute("name")+'';
			}
			if(opt.hasAttribute('name-of-file')){
				fileNames[opt.getAttribute('name-of-file')+''] = opt.getAttribute("name")+'';
			}
			if(opt.hasAttribute('stat')){
				let statType = opt.getAttribute('stat');
				let statName = opt.getAttribute('name');
				if(globalStats[statType]!==undefined) globalStats[statType]=statName;
				else statCols.push({type:statType, name:statName});
			}
			if(opt.hasAttribute('exclude-tag')){
				excludeTags.push(opt.getAttribute('exclude-tag'));
			}
			if(opt.hasAttribute('column')){
				var c = {tier:opt.getAttribute('column')+'', name:opt.hasAttribute("name") ?opt.getAttribute("name")+'':''};
				columns.push(c);
				if(c.name) tierNames[c.tier]  = c.name;
			}
		});
		var asort = (A,B)=>{ return A.start > B.start ? 1 : -1};
		var tiers = {}; // map of tier id => tier's annotations		
		var files = {}; // map of file name => file's annotations
		var groups = {}; // ... 
		var sequences = {}; // ...
		if(sequence){
			log.debug(`USING TIMELINE SEQUENCE ${sequence}`);
			group = false; // EAF based			
		}				
		if(attr.reference) tiers[attr.reference] = [];
		list.forEach(a=>{
			if(!tiers[a.tier.id]) tiers[a.tier.id] = [];
			tiers[a.tier.id].push(a);
			if(!files[a.af.id]) files[a.af.id] = [];
			files[a.af.id].push(a);
			if(!groups[a.af.group]) groups[a.af.group] = [];
			groups[a.af.group].push(a);
			if(sequence){
				a.af.getTier(sequence).annotations.forEach(seq=>{
					if(a.start>=seq.start && a.stop<=seq.stop){
						var sid = a.af.id+'/'+seq.value;
						if(!sequences[sid]){
							var s = sequences[sid] = [];
							s.id=sid;
							s.name = seq.value+'';
							s.af = seq.af;
							s.start = seq.start;
							s.stop = seq.stop;							
						}
						sequences[sid].push(a);
					}
				});
			}
		});
		var eaf, prev;
		var s = `<div class="timeflow-container">`;
        s += `<table><tbody>`;		
		var aset = group ? groups : files;
		if(sequence) {
			aset = {};
			var a = [];
			for(var k in sequences) a.push(sequences[k]);
			a.sort(
				(a,b)=>{ return a.af.filename+"" == b.af.filename+"" ?  (a.start > b.start ? 1 : -1) : (a.af.filename+"" < b.af.filename+"" ? -1 : 1)}
			);
			a.forEach(seq=>aset[seq.id]=seq);
		}
		var index=0;
		var names = group ? groupNames : fileNames;
		var rounds = [];
		var ftiers;
		var intercodedTiers = {};
		for(key in aset){
			// if there is more than one EAF/group, display a separator with its name
			var kname = sequence ? aset[key].name : names[key] || key;						
			if(Object.keys(aset).length>1) s += `<tr class="af-separator sequence-separator"><td colspan="100"><i>${esc(kname)}</i></td></tr>`;
			aset[key].sort(asort);
			var atiers = ftiers = {};
			if(columns.length) columns.forEach(col=>atiers[col.tier]=[]); // order matters
			if(attr.columns) attr.columns.split(',').forEach(col=> atiers[col] = []);			
			if(attr.reference) atiers[attr.reference] = [];
			
			var astart = {};
			aset[key].forEach(a=>{
				if(!atiers[a.tier.id]) atiers[a.tier.id] = 1;
				if(!astart[a.start]) astart[a.start] = [];
				astart[a.start].push(a);
			});
			// show tier names in header
			s += `<tr class="tier-header">`;
			if(showIndex) s+=`<td></td>`;
			if(showStart) s+=`<td></td>`;
			if(showVideo) s+=`<td></td>`;
			//if(showValue) s+=`<td></td>`;
			for(tier in atiers) s += `<td class="tier-name">`+esc(tierNames[tier] || tier)+`</td>`;			
			statCols.forEach(c=>s+=`<td class="stat-${c.type}">${esc(c.name)}</td>`);					
			s += `</tr>`;				
			
			var first = true;
			var done = [];
			var round;
			for(var start in astart){
				s += `<tr class="annotations">`;	
				
				if(showStart) s+=`<td>`+ftime(start/1000)+`</td>`;
				index++;
				round = {start:start,values:[],tierValue:{},annotations:[],tiers:{},cMatrix:{}}
				rounds.push(round);
				if(showIndex) s+=`<td>${index}</td>`;				
				if(showVideo){
					var a = astart[start][0];
					s+=`<td><video preload="none" poster="${a.snap()}" controls width="300"><source src="${a.clip()}" type="video/mp4"></video></td>`;
				}
				if(showValue){
					/*var a = astart[start][0];
					astart[start].forEach(aa=>{
						if(aa.hasTag('exclude-from-intercoding')) a = aa;
					});
					s+=`<td><span class="annotation-value">${a.value}</span></td>`;*/
				}
				Object.keys(atiers).forEach(tier=>{					
					s += `<td>`;
					astart[start].forEach(a=>{						
						if(!a.tier.id.equals(tier)) return;										
						if(a.tier.id.equals(attr.reference)){		
							round.refAnnotation = a;
							s+=`<label class="annotation annotation-reference">${esc(valueReplace(a.value))}</label>`;
						}else{
							//TODO if(a.hasTag(...) excludeTags
							var exclude = false;
							excludeTags.forEach(tag=>exclude=exclude||a.hasTag(tag));
							//if(exclude) log.info('excluded '+a.value)
							if(!exclude){
								intercodedTiers[tier] = 1;
								round.tiers[tier] = a;
								round.annotations.push(a);								
								round.values.push(valueReplace(a.value));
								round.tierValue[tier] = valueReplace(a.value);
							}
							s+=`<label class="annotation ${exclude?'excludeee':''} annotation-input">${esc(valueReplace(a.value))}</label>`;
						}
						
					});
					s += `</td>`;					
				});
				
				statCols.forEach(c=>{
					var td = '';
					var cla = '';
					switch(c.type){
						case 'score':
						case 'score/':
						case 'score%':
							var score = 0;
							var total = round.values.length;							
							var pairs = {};
							
							if(!round.refAnnotation){
								// score when no reference is set:								
								
								//td+=`reference tier missing`;								
								
								round.values.forEach((v,i)=>{
									round.values.forEach((vv,ii)=>{
										if(ii===i) return; // don't compare self
										if(!vv || !v) return; // don't compare if missing value
										if(vv===v) score++;			
									});									
								});									
							
								if(score) cla='green';
								else cla='red';
								//break;
								//TODO change total depending on what's expected
							}else{
								var ref = round.refAnnotation.value+'';
								score = round.values.length;								
								round.values.forEach(v=>{									
									if(v!=ref) score--;
								});
								if(score==total) cla='green';
								else if(score==0) cla='red';
								else cla='orange';
							}
							if(c.type.indexOf('%')>-1) td+=`${Math.round((score/total)*100)}%`;
							else if(c.type.indexOf('/')>-1) td+=`${score} / ${total}`;
							else td+=`${score}`;							
							round.score = score;
							round.totalForScore = total;
							for(let pair in pairs){
								//cMatrix[pair]+=pairs[pair];
							}
							break;
						case 'count':
							var o = {}
							round.values.forEach(v=>{
								if(!o[v]) o[v]=1;
								else o[v]++;
							});
							let keys = Object.keys(o);
							//td+=JSON.stringify(o);
							if(keys.length==1){
								cla='green';
								td+='100%';
								round.count = 100;
							}else if(keys.length==2){
								cla='orange';
								var x = o[keys[0]]>o[keys[1]] ? o[keys[0]] : o[keys[1]];
								round.count = Math.round((x/round.values.length)*100)
								td+=round.count+`%`;
								
							}else keys.forEach(k=>{
								cla='red';
								var x = Math.round((o[k]/round.values.length)*100);
								if(!round.count || round.count<x) round.count=x;
								td+=round.count+'% ';
							});
							break;
							
						default:
							td=`invalid stat: `+c.type;
					}
					s+=`<td class="stat ${cla}">${td}</td>`;
				});
				
				s+=`</tr>`;
				/*if(eaf!=a.eaf){
					eaf = a.eaf;
					s += `<tr class="eaf-separator"><td colspan="100"><i>${eaf.id}</i></td></tr>`;
					printHeaders();
				}
				if(prev == a) return; // duplicates from possible OPerations
				if(prev && (a.start - collapse*1000  > prev.stop)){
					var diff = Math.round((a.start - prev.stop)/1000);
					if(!first) s += `<tr class="time-separator"><td colspan="100"><i>${fdur(diff)}</i></td></tr>`;
				}*/
				
				

				//s += `</tr>`;
			};
		};
		if(statCols.length){
			s += `<tr class="stats">`;
			if(showIndex) s+=`<td></td>`;
			if(showStart) s+=`<td></td>`;
			if(showVideo) s+=`<td></td>`;
			//if(showValue) s+=`<td></td>`;
			for(tier in ftiers) s += `<td class="tier-stat"></td>`;			
			statCols.forEach(c=>{
					var td = '';
					var cla = '';
					switch(c.type){
						case 'score':
						case 'score/':
						case 'score%':
							//if(!attr.reference){
							//	td+=`reference tier missing`;
							//	break;
							//}
							var score = 0;
							var total = 0;
							rounds.forEach(r=>{
								score+=r.score;
								total+=r.totalForScore;
							});
							if(c.type.indexOf('%')>-1) td+=`${Math.round((score/total)*100)}%`;
							else if(c.type.indexOf('/')>-1) td+=`${score} / ${total}`;
							else td+=`${score}`;
							if(score==total) cla='green';
							else if(score==0) cla='red';
							else cla='orange';
							break;
						case 'count':
							var count = 0;
							var total = 0;
							rounds.forEach(r=>{
								count+=r.count;								
							});							
							count = Math.round(count / rounds.length);
							if(count>=90) cla='green';
							else if(count<10) cla='red';
							else cla='orange';							
							td+=count+'%';
							break;
							
						default:
							td=`invalid stat: `+c.type;
					}
					s+=`<td class="final-stat ${cla}">${td}</td>`;
				});
			s += `</tr>`;
		}
		s+=`</tbody></table></div>`;
		// COINCIDENCE MATRIX
		var cMatrix = {};
		var cReliability = {sourceUnits:0,computableUnits:0};					
		rounds.forEach(round=>{
			round.numParticipants = 0;
			round.values.forEach(v=>round.numParticipants+=(v?1:0));
			cReliability.sourceUnits++;
			if(round.numParticipants>1){
				cReliability.computableUnits++;		
				var m = {};
				round.values.forEach((v,i)=>{
					round.values.forEach((vv,ii)=>{
						if(ii===i) return; // don't compare self
						if(!vv || !v) return; // don't compare if missing value
						if(!m[v]) m[v] = {}; // initialize temp matrix
						if(!m[v][vv]) m[v][vv] = 0;
						if(!m[vv]) m[vv] = {};
						if(!m[vv][v]) m[vv][v] = 0;										
						m[v][vv]++; // increment score (formula's numerator)
						m[vv][v]++; // symmetry for easy calculations																	
					});									
				});
				for(let k in m){ // all scores must be divided
					for(let kk in m[k]){ // (formula's denominator)
						m[k][kk]/=(round.numParticipants-1);
						if(!cMatrix[k]) cMatrix[k] = {};
						if(!cMatrix[k][kk]) cMatrix[k][kk] = 0;
						cMatrix[k][kk]+=m[k][kk]; // add to main matrix											
					}
				}
			};
		});		
		var kripNumerator = 0;
		var kripDenominator = 0;
		var kripNumeratorFormula = '';
		var kripDenominatorFormula = '';
		var cFreq = {};
		var cValues = Object.keys(cMatrix).sort();
		var weightMatrix;
		if(weightMatrixFunc){
			weightMatrix = {};
			let f = new Function('return '+weightMatrixFunc)();
			var WM = '', WMCSV='(values),"'+cValues.join('","')+'"';
			WM+=`<div class="weight-matrix stat-section">`;
			WM+=`<div class="stat-title">${globalStats['weight-matrix']}</div>`;						
			WM+=`<table><thead><tr><td></td>`;
			cValues.forEach(k=>WM+=`<td>${k}</td>`);
			WM+=`</tr></thead><tbody>`;
			cValues.forEach(k=>{			
				WM+=`<tr><td>${k}</td>`;
				WMCSV+=`\n"${k}"`;
				var kNum = '';
				cValues.forEach(kk=>{
					let wv = f(k,kk);		
					if(!weightMatrix[k]) weightMatrix[k] = {};
					weightMatrix[k][kk] = wv;
					WM+=`<td>${wv}</td>`;	
					WMCSV+=`,"${wv}"`;
				});
				WM+=`</tr>`;				
			});					
			WM+=`</tr>`;		
			WM+=`</tbody></table>`;
			WM+=`</div>`;		
		}
		if(globalStats['weight-matrix']) s+=WM;
		if(saveWeightMatrixCSV) api.writeFile(saveWeightMatrixCSV, WMCSV);
		
		var CM = '', CMCSV='(values),"'+cValues.join('","')+'","(total)"\n';
		CM+=`<div class="coincidence-matrix stat-section">`;
		CM+=`<div class="stat-title">${globalStats['coincidence-matrix']}</div>`;						
		CM+=`<table><thead><tr><td></td>`;
		cValues.forEach(k=>CM+=`<td>${k}</td>`);
		CM+=`</tr></thead><tbody>`;		
		var kDone = {};
		var maxValue = 0;
		cValues.forEach(k=>{
			cValues.forEach(kk=>{
				let cm = (cMatrix[k][kk]||0)/2;
				if(maxValue<cm) maxValue = cm;
			});
		});
		cValues.forEach(k=>{
			cFreq[k] = 0;
			CM+=`<tr><td>${k}</td>`
			CMCSV+=`"${k}"`;
			var kNum = '';
			cValues.forEach(kk=>{
				let cm = (cMatrix[k][kk]||0)/2;
				let cc = 'cmvalue';
				if(k==kk) cc += ' diag';
				let level = Math.floor((cm/maxValue)*4)
				if(cm) level++;
				cc += ' cm-'+level;
				cFreq[k] += cm;
				CM+=`<td class="${cc}">${cm}</td>`;	
				CMCSV+=`,"${cm}"`;
				//https://en.wikipedia.org/wiki/Krippendorff%27s_alpha
				//only the entries in one of the off-diagonal triangles of the coincidence matrix are listed:
				let kkey = [k,kk].sort().join('*******')
				if(kDone[kkey]) return;
				kDone[kkey] = true;
				if(cm && (kk!=k)){
					kNum+=(kNum ? ' + ' : '') + cm + (useWeightMatrix && weightMatrix ? '&middot;(1-'+weightMatrix[k][kk]+')' : '');
					kripNumerator += useWeightMatrix && weightMatrix ? cm*(1-weightMatrix[k][kk]) : cm;
				}								
			});
			if(kNum) kripNumeratorFormula += (kripNumeratorFormula ? ' + ' : '') + '('+kNum+')\n';
			CM+=`<td class="frequency">${cFreq[k]}</td>`;			
			CM+=`</tr>`;
			CMCSV+=`,"${cFreq[k]}"\n`;
		});
		CM+=`<tr>`;
		CM+=`<td></td>`;
		CMCSV+=`"(total)"`;
		var cFreqTotal = 0;
		cValues.forEach(k=>{
			cFreqTotal += cFreq[k]||0;
			CM+=`<td class="frequency">${cFreq[k]}</td>`;
			CMCSV+=`,"${cFreq[k]}"`;
		});			
		
		CM+=`<td class="frequency frequency-total">${cFreqTotal}</td>`;	
		CM+=`</tr>`;		
		CM+=`</tbody></table>`;
		CMCSV+=`,"${cFreqTotal}"`;
		if(saveCoincidenceMatrixCSV) api.writeFile(saveCoincidenceMatrixCSV, CMCSV);
		//CM+=`<div class="coincidence-matrix">${JSON.stringify(cMatrix)}</div>`;		
		CM+=`<div class="coincidence-matrix-reliability">Reliability: `;			
		CM+=`<span>${cReliability.sourceUnits} units, </span>`;
		if(cReliability.sourceUnits==cReliability.computableUnits) CM+=`<span>all units are computable</span>`;
		else CM+=`<span>${cReliability.sourceUnits-cReliability.computableUnits} units (${Math.round((cReliability.sourceUnits-cReliability.computableUnits)/cReliability.sourceUnits*100)}%) are not computable because of insufficient data</span>`;
		CM+=`</div>`;		
		CM+=`</div>`;
		if(globalStats['coincidence-matrix']) s+=CM;
		
		
		
		kDone = {};
		cValues.forEach(k=>{
			var kDen='';
			cValues.forEach(kk=>{
				if(kk===k) return;
				let kkey = [k,kk].sort().join('******')
				if(kDone[kkey]) return;					
				kDone[kkey]=true; //+'['+k+'~'+kk+']'
				kDen += (kDen ? ' + ' : '' ) + cFreq[k]+'&middot;'+cFreq[kk]+(useWeightMatrix && weightMatrix ? '&middot;(1-'+weightMatrix[k][kk]+')' : '');
				kripDenominator += cFreq[k]*cFreq[kk]*(useWeightMatrix && weightMatrix ? (1-weightMatrix[k][kk]) : 1);
			});				
			if(kDen) kripDenominatorFormula += (kripDenominatorFormula ? ' + ' : '') + '('+kDen+')\n';
		});
		
		var kripFinal = 1 - (kripNumerator/((1/(cFreqTotal-1))*kripDenominator));
		var kripShort = (kripFinal+'').substr(0,5);
			
		if(saveKrippendorffAlphaJSON) api.writeFile(saveKrippendorffAlphaJSON, `{"alpha":${(kripFinal+'').substr(0,8)}}`);
		if(globalStats['krippendorff-alpha']){
			s+=`<div class="krippendorff-alpha stat-section">`;
			s+=`<div class="stat-title">${globalStats['krippendorff-alpha']}</div>`;						
				/*let adata= [];
				rounds.forEach(round=>{
					let r = [];
					Object.keys(atiers).forEach(tier=>{						
						//if(tier == attr.reference) return;
						if(!intercodedTiers[tier]) return;						
						let v = round.tierValue[tier];
						
						r.push(v===undefined ? '' : v);
					});
					adata.push(r);
					
				});*/
				s+=`<span class="krippendorff-alpha-label"><b>&alpha;</b> = </span>`;
				s+=`<span class="krippendorff-alpha-value" title="${kripFinal}">${kripShort}`;
				
				s+=`<div class="krippendorff-alpha-details">&Sigma; Numerator: ${kripNumeratorFormula} = ${roundec(kripNumerator)} <br>&Sigma; Denominator: ${kripDenominatorFormula} = ${roundec(kripDenominator)}<br>Alpha: 1 - (${roundec(kripNumerator)}/((1/(${cFreqTotal}-1))&middot;${roundec(kripDenominator)})) = ${kripFinal}</div></span>`;
				
			s+=`</div>`
		}
		return s;
	},
	css:s=>{
		return `
#view tr.time-separator {	
    height: 50px;
	background: rgb(255,255,255);
	background: linear-gradient(0deg, rgba(255,255,255,0) 0%, rgba(0,0,0,0.08167016806722693) 50%, rgba(255,255,255,0) 100%);
	text-align:center;
}
#view tr.tier-header {	
    font-weight:bold;
	text-align:center;
}
#view tr.annotation {
    vertical-align: top;
	
}
#vview tr.annotations td:first-child {
    color:gray;
	font-size: 14px;
    padding-right: 5px;
}
#view tr.af-separator td {
    background: black;
    color: white;
    font-weight: bold;
    padding: 5px;
    border-radius: 8px;
    text-align: center;
}
#view tr.annotations td:not(first-child) {
	border-right: 1px dashed gainsboro;
}
#view tr.stats td,
#view tr.annotations td {
    border-top: 1px dashed gainsboro;
    max-width: 320px;
	padding-top: 10px;
	text-align:center;

}
#view tr.stats,
#view td.stat {
	font-weight:bold;
}
#view td.final-stat {
    border-bottom: 3px solid gainsboro;
    padding-bottom: 5px;
}
#view video {
    border-radius: 6px;
	display: block;
}

#view .red {
	color:red;
}
#view .green {
	color:green;
}
#view .orange {
	color:orange;
}
#view .stat-section table {
    border-collapse: collapse;
	margin-top:10px;
}
#view .stat-section table tr td:first-child,
#view .stat-section table thead td {
	font-weight:bold;
	padding:10px
}
#view .stat-section table td {
    border: 1px solid gray;
    padding: 5px;
	text-align:center;
}
#view .stat-section table td.diag {
    border: 2px dashed gray;
    
}
#view .coincidence-matrix table tr:last-child td:first-child {
	border:none;
}
#view .coincidence-matrix table thead td:first-child {
	border:none;
}
#view .krippendorff-alpha-label,
#view .krippendorff-alpha-value,
#view .krippendorff-alpha-title,
#view .stat-section .stat-title {
    font-size: 1.5em;
	margin-top:25px;
}


#view .cm-0 { background: #fbfaea; }
#view .cm-1 { background: #fff33b; }
#view .cm-2 { background: #fdc70c; }
#view .cm-3 { background: #f3903f; }
#view .cm-4 { background: #ed683c; }
#view .cm-5 { background: #e93e3a; }

#view .stat-section {
    margin: 10px;
    padding-bottom: 5px;
}
#view .coincidence-matrix-reliability {
    margin-top: 10px;
    color: #4b4b4b;
}
#view span.krippendorff-alpha-value:hover div {
    max-height: 1000px;
}
#view .krippendorff-alpha-details {
	max-height:0px;
	overflow:hidden;
	transition:.5s all linear;
	font-size:.5em;
	white-space:pre-line;
}

`
	}
}

function fdur(s){
    var sec_num = parseInt(s, 10);
    var hours   = Math.floor(sec_num / 3600);
    var minutes = Math.floor((sec_num - (hours * 3600)) / 60);
    var seconds = sec_num - (hours * 3600) - (minutes * 60);    
	if(!hours) {
		if(!minutes) return seconds+" seconds";
		return minutes+" minutes";		
	}
    return hours+' hours';
}

function ftime(s){
    var sec_num = parseInt(s, 10);
    var hours   = Math.floor(sec_num / 3600);
    var minutes = Math.floor((sec_num - (hours * 3600)) / 60);
    var seconds = sec_num - (hours * 3600) - (minutes * 60);
    //if (hours   < 10) {hours   = "0"+hours;}
    if (minutes < 10) {minutes = "0"+minutes;}
    if (seconds < 10) {seconds = "0"+seconds;}
    return hours+':'+minutes+':'+seconds;
}

function toCSV(data){
	var json = data;
	var fields = Object.keys(json[0])
	var replacer = function(key, value) { return value === null || value === undefined ? '' : value } 
	var csv = json.map(function(row){
	return fields.map(function(fieldName){
		return JSON.stringify(row[fieldName], replacer)
	}).join(',')
	})
	csv.unshift(fields.join(',')) // add header column
	csv = csv.join('\r\n');
	return csv;
}