/** 
@author CG
@log 2024-08-16 initial implementation
@input Array<Annotation>

@attr style select (monospace) [monospace,fluid] The layout to use 
@attr separator string (: ) The separator to use between the tier name and the text
@attr max-line-width int (0) Maximum number of characters in one line of text (0 = disabled)
@attr min-tier-width int (0) Minimum width (in characters) of the tier column 
@attr export-to-csv bool (false) Whether to also export a CSV file of each transcript

Display annotations as a simple transcript, can also export transcript to CSV files.

@dependency sort-annotations.js
@dependency csv-utils.js

@sample VIEW 
@image png
**/
view.transcript = {
	render_fluid:function(list){		
		const separator = attr.separator || ': ';		
		let html = '';
		
		let current;						
		const write = (tier,text)=>{
			if(current && current.tier.id==tier){
				html+=`<span class="tier-pad"></span>`;
			}else{
				if(html) html+=`<br/>`;
				html+=`<span class="tier-name">${esc(tier)}${esc(separator)}</span>`;
			}		
			html+=`<span class="line-value">${esc(text)}</span><br/>`;			
		}		
		for(let a of list){			
			write(a.tier.id,a.value);
			current = a;
		}
		return html;
	},	
	render_monospace:function(list){		
		const separator = attr.separator || ': ';
		const lineWidth = attr['max-line-width']*1 || 0;
		const minTierWidth = attr['min-tier-width']*1 || 0;		
		const padLeft = (s,i)=>{
			while(s.length<i) s = ' '+s;
			return s;
		}
		let html = '';
		let tiers = {};
		let current;
		for(let a of list) if(!tiers[a.tier.id]) tiers[a.tier.id]=1;
		let maxTierLength = 0;
		Object.keys(tiers).forEach(t=>{
			if(t.length>maxTierLength) maxTierLength=t.length;
		});
		if(maxTierLength<minTierWidth) maxTierLength=minTierWidth;
		//log.info(`Max Tier Name Length: ${maxTierLength}, Line Width: ${lineWidth}`);
		const tierPad = maxTierLength+separator.length;
		const write = (tier,text)=>{
			if(current && current.tier.id==tier){
				html+=`<span class="tier-pad">${padLeft('',tierPad)}</span>`;
			}else{
				html+=`<span class="tier-name">${esc(padLeft(tier,maxTierLength))}${esc(separator)}</span>`;
			}
			var first = true;
			const outputLine = line =>{
				if(!first) html+=`<span class="tier-pad">${padLeft('',tierPad)}</span>`;
						html+=`<span class="line-value">${esc(line)}\n</span>`;
						first = false;
			}
			let lines = text.split("\n")
			let line = '';
			
			for(let aLine of lines){
				line='';
				let words = aLine.split(' ');
								
				for(let w of words){
					if(lineWidth && (line.length + w.length + 1 > lineWidth)){
						outputLine(line);
						line = w;					
					}else{
						line+= line.length ? ' '+w : w;
					}
				}
			
				if(line) outputLine(line);
			}			
		}		
		for(let a of list){			
			write(a.tier.id,a.value);
			current = a;
		}
		return html;
	},
	list:function(list){
		const exportToCSV = attr['export-to-csv']=='true';
		let afList = {};
		for(let a of list){
			if(!a || !a.get) continue;			
			if(!afList[a.af]) afList[a.af] = [];
			afList[a.af].af = a.af;
			afList[a.af].push(a);
		}	
		
		const style = attr.style || 'monospace';
		let html = `<div class="transcript-${style}">`;		
		for(let af in afList){
			let annotations = sortAnnotations(afList[af]);
			if(exportToCSV) this.exportToCSV(afList[af].af.id+'.csv',annotations);
			html += `<div class="transcript-file">`;
			html += `<div class="transcript-title">${afList[af].af.file.getName()}</div>`;
			html += `<div class="transcript-content">`;
			html += this['render_'+style](annotations);
			html += `</div></div>`
		}
		html+=`</div>`;		
		return html;
	},
	exportToCSV:function(csvid,list){
		try {
			let data = [];
			for(let a of list){
				data.push({
					tier:a.tier.id,
					value:a.value
				});
			}
			let c = api.writeFile(csvid, arrayToCSV(data));
			log.info(`Saved CSV file "${csvid}"`);
		}catch(ex){
			log.warn(`FAILED TO WRITE FILE: "${csvid}"`);
			log.error(ex.message);
		}
	},
	css:s=>{
		return `
		#view .transcript-monospace .transcript-file {
			
		}
		#view .transcript-monospace .transcript-content {
			border-left:2px solid gainsboro;
			padding-left:20px
		}
		#view .transcript-monospace .transcript-file:not(:last-child) {
			margin-bottom:30px
		}
		#view .transcript-monospace .transcript-title {			
			margin-bottom: 10px;
			font-weight: bold;
		}
		#view .transcript-monospace {
			white-space: pre;
			font-family: monospace;
		}		
		
		#view .transcript-fluid .transcript-file {
			
		}
		#view .transcript-fluid .tier-name {
			font-style:italic;
		}
		#view .transcript-fluid .transcript-content {
			border-left:2px solid gainsboro;
			padding-left:20px
		}
		#view .transcript-fluid .transcript-file:not(:last-child) {
			margin-bottom:30px
		}
		#view .transcript-fluid .transcript-title {			
			margin-bottom: 10px;
			font-weight: bold;
		}
		#view .transcript-fluid {
			white-space: wordwrap;			
		}
`;
	}
}