/** 
@author CG
@license Wordcloud2 by Timothy Guan-tin Chien and contributors https://github.com/timdream/wordcloud2.js | MIT
@log 2024-03-24 added threshold attributes + size-factor auto
@input Array<Annotation>
@input Array<Array> an array of ["word", weight] pairs
@todo input String a text to take words from and calculate frequency
@attr length-threshold int (3) Exclude words with less than length-threshold letters
@attr frequency-threshold int (0) Exclude words with a lower frequency 
@attr size-factor int (0) The size factor to draw words (0 = auto) 
@attr min-size int (0) (0 to disable)
@attr weight-factor number (1)
@attr font-family string
@attr font-weight string (normal)
@attr color string (random-dark)
@attr background-color string (#fff)
@attr draw-out-of-bound bool (false)
@attr shrink-to-fit bool (false)
@attr origin array origin of the “cloud” in [x, y]
@attr min-rotation number (-Math.PI / 2)
@attr max-rotation number (Math.PI / 2)
@attr rotation-steps number (0)
@attr rotate-ratio number (0.1)
@attr shape string (circle) [circle,cardioid,diamond,square,triangle-forward,triangle,pentagon,star]
@attr ellipticity number (0.65)
@attr shuffle bool (true)
@attr list-words int (0) How many words to show in a table, 0 to disable

@todo clear-canvas bool true
@todo grid-size int 8
@todo drawMask: false, // visualize the grid
@todo maskColor: 'rgba(255,0,0,0.3)',
@todo maskGapWidth: 0.3,
@todo wait: 0,
@todo abortThreshold: 0, // disabled
@todo abort: function noop () {},
@todo classes string (null)

The wordcloud helps visualize frequency of words and has many customisation attributes.
Wordcloud can slow down PDF generation time and can take some time to show-up.

@image png
**/
view.wordcloud = {
	includes:['wordcloud2.js'],
	data:data=>{		
		var listWordsCount = attr['list-words']*1||0;
		var frequencyThreshold = attr['frequency-threshold']*1||0;		
		var lengthThreshold = attr['length-threshold']*1||0;
		if(attr['length-threshold']===null){ // because 0 should be a valid threshold?
			lengthThreshold = 3;
		}		
		var sizeFactor = attr['size-factor']*1||0;
		var words = {};
		if(!data.length) return ``;
		//var t = typeof data;		
		if(false){
			data.split(' ').forEach(w=>{
				w = w.trim();
				if(!w) return;
				if(!words[w]) words[w]=0;
				words[w]++
			});
			var weights = [];
			for(var k in words) if(k.length>3) weights.push([k,words[k]]);
			words = JSON.stringify(weights);
		}else{
			var first = data[0];
			if(first.length==2){ // data is probably an array of weighted words
				words = JSON.stringify(data);;
			}else if(first.clip){ // data is a list of annotations			
				data.forEach(a=>{
					var v = (a.value || '') + '';
					v.split(' ').forEach(w=>{
						w = w.trim();
						if(!w) return;
						if(!words[w]) words[w]=0;
						words[w]++
					});
				});
				var weights = [];
				var failed = [];
				var exmsg = '';
				var threshFreqTotal = 0 ;
				var threshLengthTotal = 0;
				for(var k in words) if(k.length>=lengthThreshold) {
					let freq = words[k];
					if(freq<frequencyThreshold){
						threshFreqTotal++;
						continue;
					}
					try {
						var b64test = btoa(k);						
						weights.push([k,words[k]]);
					}catch(ex){
						failed.push(k);
						exmsg = ex.message;
						//log.warn("Failed encoding: "+k);
					}					
				}else{
					threshLengthTotal++;
				}
				words = JSON.stringify(weights);		
				if(failed.length) {
					log.warn(failed.length+" words could not be encoded");
					log.report({
						error:exmsg,
						title:'words could not be encoded to base64',
						words:failed
					});
				}
			}
		}		
		var o = {};
		if(!sizeFactor) {
			attr['size-factor'] = Math.round(weights.length/5);
			log.info(`Using best detected size-factor of ${attr['size-factor']}`)
		}
		for(var k in attr){
			if(k=='learn' || k=='type' || k=='list-words') continue;
			var p = snakeToCamel(k);
			var v = o[p] = attr[k]+'';			
			if(v==='true') o[p]=true;
			else if(v==='false') o[p]=false;
			else if(!isNaN(v*1)) o[p]=v*1; // convert relevant string values to number
		}		
		var bbb = btoa(words);
		log.info(`${weights.length} unique words`);
		if(threshFreqTotal){
			log.info(`${threshFreqTotal} words excluded because below frequency threshold of ${frequencyThreshold}`);
		}
		if(threshLengthTotal){
			log.info(`${threshLengthTotal} words excluded because less than ${lengthThreshold} letters`);
		}
		bbb = weights.sort((a,b)=> a[1]>b[1] ? -1 : 1);
		var html = `<canvas width="700" height="500" class="wordcloud" data-opt="${btoa(JSON.stringify(o))}" data-words="${btoa(words)}"></canvas>`;	
		if(listWordsCount){
			html+=`<table class="words-list">`;
			//html+=`<thead><tr><td>Word</td><td>Frequency</td></tr><tbody>`;
			html+=`<thead><tr><td>Word</td><td>Frequency</td></tr><tbody>`;
			for(let i =0;i<listWordsCount;i++){
				html+=`<tr><td>${bbb[i][0]}</td><td><i>&times;</i><span>${bbb[i][1]}<span></td></tr>`;
			}
			//s+=``
			//words.for
			html+=`</tbody></table>`;
		}			
		return html;
	},
	css:s=>{
		return ``;		
	},
	js:s=>{
		$(ready=>{
			if(!window.WordCloud){
				console.warn('wordcloud2.js is not loaded, impossible to display wordcloud');
				return;
			}
			$('.view-wordcloud .wordcloud').each((idx,e)=>{
				var opt = JSON.parse(atob($(e).data('opt'))) || {};
				opt.list = JSON.parse(atob($(e).data('words')));
				opt.weightFactor = size => size * e.width / (opt.sizeFactor || 10);				
				//console.info('WC',e,opt);
				WordCloud(e, opt);
			});
		});
	}
}

function snakeToCamel(str){
  return str.toLowerCase().replace(/([-_][a-z])/g, group =>
    group
      .toUpperCase()
      .replace('-', '')
      .replace('_', '')
  );
}